# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Values come from the Redfish System json-schema 1.0.0:
# http://redfish.dmtf.org/schemas/v1/ComputerSystem.v1_0_0.json#/definitions/ComputerSystem  # noqa

from sushy.resources import constants as res_cons

# Reset action constants

RESET_ON = res_cons.RESET_TYPE_ON
RESET_FORCE_OFF = res_cons.RESET_TYPE_FORCE_OFF
RESET_GRACEFUL_SHUTDOWN = res_cons.RESET_TYPE_GRACEFUL_SHUTDOWN
RESET_GRACEFUL_RESTART = res_cons.RESET_TYPE_GRACEFUL_RESTART
RESET_FORCE_RESTART = res_cons.RESET_TYPE_FORCE_RESTART
RESET_NMI = res_cons.RESET_TYPE_NMI
RESET_FORCE_ON = res_cons.RESET_TYPE_FORCE_ON
RESET_PUSH_POWER_BUTTON = res_cons.RESET_TYPE_PUSH_POWER_BUTTON

# System' PowerState constants

SYSTEM_POWER_STATE_ON = res_cons.POWER_STATE_ON
"""The system is powered on"""

SYSTEM_POWER_STATE_OFF = res_cons.POWER_STATE_OFF
"""The system is powered off, although some components may continue to
   have AUX power such as management controller"""

SYSTEM_POWER_STATE_POWERING_ON = res_cons.POWER_STATE_POWERING_ON
"""A temporary state between Off and On. This temporary state can
   be very short"""

SYSTEM_POWER_STATE_POWERING_OFF = res_cons.POWER_STATE_POWERING_OFF
"""A temporary state between On and Off. The power off action can take
   time while the OS is in the shutdown process"""

# Indicator LED Constants

SYSTEM_INDICATOR_LED_LIT = res_cons.INDICATOR_LED_LIT
"""The Indicator LED is lit

Deprecated: Use `sushy.resources.constants.INDICATOR_LED_LIT`.
"""

SYSTEM_INDICATOR_LED_BLINKING = res_cons.INDICATOR_LED_BLINKING
"""The Indicator LED is blinking

Deprecated: Use `sushy.resources.constants.INDICATOR_LED_BLINKING`.
"""

SYSTEM_INDICATOR_LED_OFF = res_cons.INDICATOR_LED_OFF
"""The Indicator LED is off

Deprecated: Use `sushy.resources.constants.INDICATOR_LED_OFF`.
"""

SYSTEM_INDICATOR_LED_UNKNOWN = res_cons.INDICATOR_LED_UNKNOWN
"""The state of the Indicator LED cannot be determine

Deprecated: Use `sushy.resources.constants.INDICATOR_LED_UNKNOWN`.
"""

# Boot source target constants

BOOT_SOURCE_TARGET_NONE = 'none'
"""Boot from the normal boot device"""

BOOT_SOURCE_TARGET_PXE = 'pxe'
"""Boot from the Pre-Boot EXecution (PXE) environment"""

BOOT_SOURCE_TARGET_FLOPPY = 'floppy'
"""Boot from the floppy disk drive"""

BOOT_SOURCE_TARGET_CD = 'cd'
"""Boot from the CD/DVD disc"""

BOOT_SOURCE_TARGET_USB = 'usb'
"""Boot from a USB device as specified by the system BIOS"""

BOOT_SOURCE_TARGET_HDD = 'hdd'
"""Boot from a hard drive"""

BOOT_SOURCE_TARGET_BIOS_SETUP = 'bios setup'
"""Boot to the BIOS Setup Utility"""

BOOT_SOURCE_TARGET_UTILITIES = 'utilities'
"""Boot the manufacturer's Utilities program(s)"""

BOOT_SOURCE_TARGET_DIAGS = 'diags'
"""Boot the manufacturer's Diagnostics program"""

BOOT_SOURCE_TARGET_SD_CARD = 'sd card'
"""Boot from an SD Card"""

BOOT_SOURCE_TARGET_UEFI_TARGET = 'uefi target'
"""Boot to the UEFI Device specified in the
   UefiTargetBootSourceOverride property"""

BOOT_SOURCE_TARGET_UEFI_SHELL = 'uefi shell'
"""Boot to the UEFI Shell"""

BOOT_SOURCE_TARGET_UEFI_HTTP = 'uefi http'
"""Boot from a UEFI HTTP network location"""

BOOT_SOURCE_TARGET_USB_CD = 'usb cd'
"""Boot from a USB CD device as specified by the system BIOS.
   NOTE(janders): This is NOT a standard value.
   On SuperMicro X11 and X12 machines, virtual media is presented as an USB CD
   drive as opposed to a CD drive. Both are present in the list of boot
   devices, however only selecting UsbCd as the boot source results in a
   successful boot from vMedia. If CD is selected, boot fails even if vMedia
   is inserted."""

# Boot source mode constants

BOOT_SOURCE_MODE_BIOS = 'bios'
BOOT_SOURCE_MODE_UEFI = 'uefi'

# Boot source enabled constants

BOOT_SOURCE_ENABLED_ONCE = 'once'
BOOT_SOURCE_ENABLED_CONTINUOUS = 'continuous'
BOOT_SOURCE_ENABLED_DISABLED = 'disabled'

# Processor related constants
# Values comes from the Redfish Processor json-schema 1.3.0:
# http://redfish.dmtf.org/schemas/v1/Processor.v1_3_0.json

# Processor Architecture constants

PROCESSOR_ARCH_x86 = 'x86 or x86-64'
PROCESSOR_ARCH_IA_64 = 'Intel Itanium'
PROCESSOR_ARCH_ARM = 'ARM'
PROCESSOR_ARCH_MIPS = 'MIPS'
PROCESSOR_ARCH_OEM = 'OEM-defined'

# Processor type constants

PROCESSOR_TYPE_ACCELERATOR = 'An Accelerator'
PROCESSOR_TYPE_CPU = 'A Central Processing Unit'
PROCESSOR_TYPE_CORE = 'A Core in a Processor'
PROCESSOR_TYPE_DSP = 'A Digital Signal Processor'
PROCESSOR_TYPE_FPGA = 'A Field Programmable Gate Array'
PROCESSOR_TYPE_GPU = 'A Graphics Processing Unit'
PROCESSOR_TYPE_OEM = 'An OEM-defined Processing Unit'
PROCESSOR_TYPE_THREAD = 'A Thread in a Processor'

# Processor InstructionSet constants

PROCESSOR_INSTRUCTIONSET_ARM_A32 = 'ARM 32-bit'
PROCESSOR_INSTRUCTIONSET_ARM_A64 = 'ARM 64-bit'
PROCESSOR_INSTRUCTIONSET_IA_64 = 'Intel IA-64'
PROCESSOR_INSTRUCTIONSET_MIPS32 = 'MIPS 32-bit'
PROCESSOR_INSTRUCTIONSET_MIPS64 = 'MIPS 64-bit'
PROCESSOR_INSTRUCTIONSET_OEM = 'OEM-defined'
PROCESSOR_INSTRUCTIONSET_x86 = 'x86 32-bit'
PROCESSOR_INSTRUCTIONSET_x86_64 = 'x86 64-bit'

# System type constants

SYSTEM_TYPE_PHYSICAL = "Physical"
"""A physical computer system"""
SYSTEM_TYPE_VIRTUAL = "Virtual"
"""A virtual machine instance"""
SYSTEM_TYPE_OS = "OS"
"""An operating system instance"""
SYSTEM_TYPE_PHYSICALLY_PARTITIONED = "PhysicallyPartitioned"
"""A hardware-based partition of a computer system"""
SYSTEM_TYPE_VIRTUALLY_PARTITIONED = "VirtuallyPartitioned"
"""A virtual or software-based partition of a computer system"""
SYSTEM_TYPE_COMPOSED = "Composed"
"""A computer system created by binding resource blocks together"""

# Secure boot constants

SECURE_BOOT_ENABLED = "Enabled"
"""UEFI secure boot is enabled."""

SECURE_BOOT_DISABLED = "Disabled"
"""UEFI secure boot is disabled."""

SECURE_BOOT_MODE_SETUP = "SetupMode"
SECURE_BOOT_MODE_USER = "UserMode"
SECURE_BOOT_MODE_AUDIT = "AuditMode"
SECURE_BOOT_MODE_DEPLOYED = "DeployedMode"

SECURE_BOOT_RESET_KEYS_TO_DEFAULT = "ResetAllKeysToDefault"
SECURE_BOOT_RESET_KEYS_DELETE_ALL = "DeleteAllKeys"
SECURE_BOOT_RESET_KEYS_DELETE_PK = "DeletePK"

SECURE_BOOT_PLATFORM_KEY = "PK"
SECURE_BOOT_KEY_EXCHANGE_KEYS = "KEK"
SECURE_BOOT_ALLOWED_KEYS_DATABASE = "db"
SECURE_BOOT_DENIED_KEYS_DATABASE = "dbx"
SECURE_BOOT_RECOVERY_KEYS_DATABASE = "dbr"
SECURE_BOOT_TIMESTAMP_DATABASE = "dbt"

SECURE_BOOT_DEFAULT_PLATFORM_KEY = "PKDefault"
SECURE_BOOT_DEFAULT_KEY_EXCHANGE_KEYS = "KEKDefault"
SECURE_BOOT_DEFAULT_ALLOWED_KEYS_DATABASE = "dbDefault"
SECURE_BOOT_DEFAULT_DENIED_KEYS_DATABASE = "dbxDefault"
SECURE_BOOT_DEFAULT_RECOVERY_KEYS_DATABASE = "dbrDefault"
SECURE_BOOT_DEFAULT_TIMESTAMP_DATABASE = "dbtDefault"



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# This is referred from Redfish standard schema.
# http://redfish.dmtf.org/schemas/v1/Thermal.v1_3_0.json

from sushy.resources import base
from sushy.resources.chassis.thermal import mappings as the_maps
from sushy.resources import common
from sushy.resources import mappings as res_maps
from sushy import utils


class Sensor(base.ListField):
    """The sensor device/s associated with Thermal."""

    identity = base.Field('MemberId', required=True)
    """Identifier of the Sensor"""

    lower_threshold_critical = base.Field('LowerThresholdCritical',
                                          adapter=utils.int_or_none)
    """Below normal range but not yet fatal"""

    lower_threshold_fatal = base.Field('LowerThresholdFatal',
                                       adapter=utils.int_or_none)
    """Below normal range and is fatal"""

    lower_threshold_non_critical = base.Field('LowerThresholdNonCritical',
                                              adapter=utils.int_or_none)
    """Below normal range"""

    name = base.Field('Name')
    """The name of this sensor"""

    physical_context = base.Field('PhysicalContext')
    """Area or device associated with this sensor"""

    status = common.StatusField('Status')
    """Status of the sensor"""

    upper_threshold_critical = base.Field('UpperThresholdCritical',
                                          adapter=utils.int_or_none)
    """Above normal range but not yet fatal"""

    upper_threshold_fatal = base.Field('UpperThresholdFatal',
                                       adapter=utils.int_or_none)
    """Above normal range and is fatal"""

    upper_threshold_non_critical = base.Field('UpperThresholdNonCritical',
                                              adapter=utils.int_or_none)
    """Above normal range"""


class FansListField(Sensor):
    """The Fan device/s associated with Thermal."""

    indicator_led = base.MappedField('IndicatorLED',
                                     res_maps.INDICATOR_LED_VALUE_MAP)
    """The state of the indicator LED, used to identify the fan"""

    manufacturer = base.Field('Manufacturer')
    """This is the manufacturer of this Fan"""

    max_reading_range = base.Field('MaxReadingRange',
                                   adapter=utils.int_or_none)
    """Maximum value for Reading"""

    min_reading_range = base.Field('MinReadingRange',
                                   adapter=utils.int_or_none)
    """Minimum value for Reading"""

    model = base.Field('Model')
    """The model of this Fan"""

    part_number = base.Field('PartNumber')
    """Part number of this Fan"""

    reading = base.Field('Reading', adapter=utils.int_or_none)
    """Current Fan Speed"""

    reading_units = base.MappedField('ReadingUnits',
                                     the_maps.FAN_READING_UNITS_MAP)
    """Units in which the reading and thresholds are measured"""

    serial_number = base.Field('SerialNumber')
    """Serial number of this Fan"""


class TemperaturesListField(Sensor):
    """The Temperature device/s associated with Thermal."""

    max_allowable_operating_value = base.Field('MaxAllowableOperatingValue',
                                               adapter=utils.int_or_none)
    """Maximum allowable operating temperature for this equipment"""

    min_allowable_operating_value = base.Field('MinAllowableOperatingValue',
                                               adapter=utils.int_or_none)
    """Minimum allowable operating temperature for this equipment"""

    max_reading_range_temp = base.Field('MaxReadingRangeTemp')
    """Maximum value for ReadingCelsius"""

    min_reading_range_temp = base.Field('MinReadingRangeTemp')
    """Minimum value for ReadingCelsius"""

    reading_celsius = base.Field('ReadingCelsius')
    """Temperature"""

    sensor_number = base.Field('SensorNumber', adapter=utils.int_or_none)
    """A numerical identifier to represent the temperature sensor"""


class Thermal(base.ResourceBase):
    """This class represents a Thermal resource."""

    identity = base.Field('Id')
    """Identifier of the resource"""

    name = base.Field('Name')
    """The name of the resource"""

    status = common.StatusField('Status')
    """Status of the resource"""

    fans = FansListField('Fans', default=[])
    """A tuple of Fan identities"""

    temperatures = TemperaturesListField('Temperatures', default=[])
    """A tuple of Temperature identities"""



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/Processor.v1_3_0.json

import collections
import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources import common
from sushy.resources.system import mappings as sys_maps
from sushy import utils

# Representation of Summary of Processor information
ProcessorSummary = collections.namedtuple('ProcessorSummary',
                                          ['count', 'architecture'])
LOG = logging.getLogger(__name__)


class ProcessorIdField(base.CompositeField):

    effective_family = base.Field('EffectiveFamily')
    """The processor effective family"""

    effective_model = base.Field('EffectiveModel')
    """The processor effective model"""

    identification_registers = base.Field('IdentificationRegisters')
    """The processor identification registers"""

    microcode_info = base.Field('MicrocodeInfo')
    """The processor microcode info"""

    step = base.Field('Step')
    """The processor stepping"""

    vendor_id = base.Field('VendorID')
    """The processor vendor id"""


class Processor(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The processor identity string"""

    socket = base.Field('Socket')
    """The socket or location of the processor"""

    processor_type = base.MappedField(
        'ProcessorType', sys_maps.PROCESSOR_TYPE_VALUE_MAP)
    """The type of processor"""

    processor_architecture = base.MappedField(
        'ProcessorArchitecture', sys_maps.PROCESSOR_ARCH_VALUE_MAP)
    """The architecture of the processor"""

    instruction_set = base.MappedField(
        'InstructionSet', sys_maps.PROCESSOR_INSTRUCTIONSET_VALUE_MAP)
    """The instruction set of the processor"""

    manufacturer = base.Field('Manufacturer')
    """The processor manufacturer"""

    model = base.Field('Model')
    """The product model number of this device"""

    max_speed_mhz = base.Field('MaxSpeedMHz', adapter=utils.int_or_none)
    """The maximum clock speed of the processor in MHz."""

    processor_id = ProcessorIdField('ProcessorId')
    """The processor id"""

    status = common.StatusField('Status')
    """The processor status"""

    total_cores = base.Field('TotalCores', adapter=utils.int_or_none)
    """The total number of cores contained in this processor"""

    total_threads = base.Field('TotalThreads', adapter=utils.int_or_none)
    """The total number of execution threads supported by this processor"""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a Processor

        :param connector: A Connector instance
        :param identity: The identity of the processor
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        """
        super(Processor, self).__init__(
            connector, identity, redfish_version, registries)

    def _get_subprocessor_collection_path(self):
        """Helper function to find the SubProcessors path"""
        subproc_col = self.json.get('SubProcessors')
        if not subproc_col:
            raise exceptions.MissingAttributeError(
                attribute='SubProcessors', resource=self._path)
        return subproc_col.get('@odata.id')

    @property
    @utils.cache_it
    def sub_processors(self):
        """A reference to the collection of Sub-Processors"""
        return ProcessorCollection(
            self._conn, self._get_subprocessor_collection_path(),
            redfish_version=self.redfish_version)


class ProcessorCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return Processor

    @property
    @utils.cache_it
    def summary(self):
        """Property to provide ProcessorSummary info

        It is calculated once when the first time it is queried. On refresh,
        this property gets reset.

        :returns: A namedtuple containing the ``count`` of processors
            in regards to logical CPUs, and their ``architecture``.
        """
        count, architecture = 0, None
        for proc in self.get_members():
            # Note(deray): It attempts to detect the number of CPU cores.
            # It returns the number of logical CPUs.
            if proc.total_threads is not None:
                count += proc.total_threads

            # Note(deray): Bail out of checking the architecture info
            # if you have already got hold of any one of the processors'
            # architecture information.
            if (architecture is None
                    and proc.processor_architecture is not None):
                architecture = proc.processor_architecture

        return ProcessorSummary(count=count, architecture=architecture)

    def __init__(self, connector, path, redfish_version=None, registries=None):
        """A class representing a ProcessorCollection

        :param connector: A Connector instance
        :param path: The canonical path to the Processor collection resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        """
        super(ProcessorCollection, self).__init__(
            connector, path, redfish_version, registries)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Values comes from the Redfish System json-schema 1.0.0:
# http://redfish.dmtf.org/schemas/v1/Resource.json or
# https://redfish.dmtf.org/schemas/v1/MessageRegistry.v1_1_1.json

# Health related constants.
HEALTH_OK = 'ok'
HEALTH_WARNING = 'warning'
HEALTH_CRITICAL = 'critical'

# State related constants.
STATE_ENABLED = 'enabled'
STATE_DISABLED = 'disabled'
STATE_STANDBYOFFLINE = 'standby offline'
STATE_STANDBYSPARE = 'standby spare'
STATE_INTEST = 'in test'
STATE_STARTING = 'starting'
STATE_ABSENT = 'absent'
STATE_UNAVAILABLEOFFLINE = 'unavailable offline'
STATE_DEFERRING = 'deferring'
STATE_QUIESCED = 'quiesced'
STATE_UPDATING = 'updating'

# Task state related constants
TASK_STATE_NEW = 'new'
TASK_STATE_STARTING = 'starting'
TASK_STATE_RUNNING = 'running'
TASK_STATE_SUSPENDED = 'suspended'
TASK_STATE_INTERRUPTED = 'interrupted'
TASK_STATE_PENDING = 'pending'
TASK_STATE_STOPPING = 'stopping'
TASK_STATE_COMPLETED = 'completed'
TASK_STATE_KILLED = 'killed'
TASK_STATE_EXCEPTION = 'exception'
TASK_STATE_SERVICE = 'service'
TASK_STATE_CANCELLING = 'cancelling'
TASK_STATE_CANCELLED = 'cancelled'

# Message Registry message parameter type related constants.
PARAMTYPE_STRING = 'string'
PARAMTYPE_NUMBER = 'number'

# Severity related constants
SEVERITY_OK = 'ok'
SEVERITY_WARNING = 'warning'
SEVERITY_CRITICAL = 'critical'

# Indicator LED Constants

INDICATOR_LED_LIT = 'indicator led lit'
"""The Indicator LED is lit"""

INDICATOR_LED_BLINKING = 'indicator led blinking'
"""The Indicator LED is blinking"""

INDICATOR_LED_OFF = 'indicator led off'
"""The Indicator LED is off"""

INDICATOR_LED_UNKNOWN = 'indicator led unknown'
"""The state of the Indicator LED cannot be determine"""

# System' PowerState constants

POWER_STATE_ON = 'on'
"""The resource is powered on"""

POWER_STATE_OFF = 'off'
"""The resource is powered off, although some components may continue to
   have AUX power such as management controller"""

POWER_STATE_POWERING_ON = 'powering on'
"""A temporary state between Off and On. This temporary state can
   be very short"""

POWER_STATE_POWERING_OFF = 'powering off'
"""A temporary state between On and Off. The power off action can take
   time while the OS is in the shutdown process"""

# Reset action constants

RESET_TYPE_ON = 'on'
"""Turn the unit on"""

RESET_TYPE_FORCE_ON = 'force on'
"""Turn the unit on immediately"""

RESET_TYPE_FORCE_OFF = 'force off'
"""Turn the unit off immediately (non-graceful shutdown)"""

RESET_TYPE_GRACEFUL_SHUTDOWN = 'graceful shutdown'
"""Perform a graceful shutdown and power off"""

RESET_TYPE_GRACEFUL_RESTART = 'graceful restart'
"""Perform a graceful shutdown followed by a restart of the system"""

RESET_TYPE_FORCE_RESTART = 'force restart'
"""Perform an immediate (non-graceful) shutdown, followed by a restart"""

RESET_TYPE_NMI = 'nmi'
"""Generate a Diagnostic Interrupt (usually an NMI on x86 systems) to cease
normal operations, perform diagnostic actions and typically halt the system"""

RESET_TYPE_PUSH_POWER_BUTTON = 'push power button'
"""Simulate the pressing of the physical power button on this unit"""

RESET_TYPE_POWER_CYCLE = 'power cycle'
"""Perform a power cycle of the unit"""

# Protocol type constants

PROTOCOL_TYPE_AHCI = 'Advanced Host Controller Interface'
PROTOCOL_TYPE_CIFS = 'Common Internet File System Protocol'
PROTOCOL_TYPE_FC = 'Fibre Channel'
PROTOCOL_TYPE_FCP = 'Fibre Channel Protocol for SCSI'
PROTOCOL_TYPE_FCoE = 'Fibre Channel over Ethernet'
PROTOCOL_TYPE_FICON = 'FIbre CONnection (FICON)'
PROTOCOL_TYPE_FTP = 'File Transfer Protocol'
PROTOCOL_TYPE_HTTP = 'Hypertext Transport Protocol'
PROTOCOL_TYPE_HTTPS = 'Secure Hypertext Transport Protocol'
PROTOCOL_TYPE_I2C = 'Inter-Integrated Circuit Bus'
PROTOCOL_TYPE_NFS = 'Network File System Protocol'
PROTOCOL_TYPE_NFSv3 = 'Network File System version 3'
PROTOCOL_TYPE_NFSv4 = 'Network File System version 4'
PROTOCOL_TYPE_NVMe = 'Non-Volatile Memory Express'
PROTOCOL_TYPE_NVMeOverFabrics = 'NVMe over Fabrics'
PROTOCOL_TYPE_OEM = 'OEM specific'
PROTOCOL_TYPE_PCIe = 'PCI Express'
PROTOCOL_TYPE_RoCE = 'RDMA over Converged Ethernet Protocol'
PROTOCOL_TYPE_RoCEv2 = 'RDMA over Converged Ethernet Protocol Version 2'
PROTOCOL_TYPE_SAS = 'Serial Attached SCSI'
PROTOCOL_TYPE_SATA = 'Serial AT Attachment'
PROTOCOL_TYPE_SCP = 'Secure File Copy Protocol'
PROTOCOL_TYPE_SFTP = 'Secure File Transfer Protocol'
PROTOCOL_TYPE_SMB = 'Server Message Block (CIFS Common Internet File System)'
PROTOCOL_TYPE_TFTP = 'Trivial File Transfer Protocol'
PROTOCOL_TYPE_UHCI = 'Universal Host Controller Interface'
PROTOCOL_TYPE_USB = 'Universal Serial Bus'
PROTOCOL_TYPE_iSCSI = 'Internet SCSI'
PROTOCOL_TYPE_iWARP = 'Internet Wide Area Remote Direct Memory Access Protocol'

# Durable name format constants

DURABLE_NAME_FORMAT_EUI = 'IEEE-defined 64-bit Extended Unique Identifier'
DURABLE_NAME_FORMAT_FC_WWN = 'Fibre Channel World Wide Name'
DURABLE_NAME_FORMAT_NAA = 'Name Address Authority Format'
DURABLE_NAME_FORMAT_NQN = 'NVMe Qualified Name'
DURABLE_NAME_FORMAT_NSID = 'NVM Namespace Identifier'
DURABLE_NAME_FORMAT_UUID = 'Universally Unique Identifier'
DURABLE_NAME_FORMAT_iQN = 'iSCSI Qualified Name'

# Apply time constants

APPLY_TIME_IMMEDIATE = 'immediate'
APPLY_TIME_ON_RESET = 'on reset'
APPLY_TIME_MAINT_START = 'at maintenance window start'
APPLY_TIME_MAINT_RESET = 'in maintenance window on reset'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values comes from the Redfish Chassis json-schema 1.8.0:
# http://redfish.dmtf.org/schemas/v1/Chassis.v1_8_0.json#/definitions/Chassis

# Chassis Types constants

CHASSIS_TYPE_RACK = 'rack chassis type'
"""An equipment rack, typically a 19-inch wide freestanding unit"""

CHASSIS_TYPE_BLADE = 'blade chassis type'
"""Blade

An enclosed or semi-enclosed, typically vertically-oriented, system
chassis which must be plugged into a multi-system chassis to function
normally.
"""

CHASSIS_TYPE_ENCLOSURE = 'enclosure chassis type'
"""A generic term for a chassis that does not fit any other description"""

CHASSIS_TYPE_STAND_ALONE = 'stand alone chassis type'
"""StandAlone

A single, free-standing system, commonly called a tower or desktop
chassis.
"""

CHASSIS_TYPE_RACK_MOUNT = 'rack mount chassis type'
"""RackMount

A single system chassis designed specifically for mounting in an
equipment rack.
"""

CHASSIS_TYPE_CARD = 'card chassis type'
"""Card

A loose device or circuit board intended to be installed in a system or
other enclosure.
"""

CHASSIS_TYPE_CARTRIDGE = 'cartridge chassis type'
"""Cartridge

A small self-contained system intended to be plugged into a multi-system
chassis"""

CHASSIS_TYPE_ROW = 'row chassis type'
"""A collection of equipment rack"""

CHASSIS_TYPE_POD = 'pod chassis type'
"""Pod

A collection of equipment racks in a large, likely transportable,
container"""

CHASSIS_TYPE_EXPANSION = 'expansion chassis type'
"""A chassis which expands the capabilities or capacity of another chassis"""

CHASSIS_TYPE_SIDECAR = 'sidecar chassis type'
"""Sidecar

A chassis that mates mechanically with another chassis to expand its
capabilities or capacity.
"""

CHASSIS_TYPE_ZONE = 'zone chassis type'
"""Zone

A logical division or portion of a physical chassis that contains multiple
devices or systems that cannot be physically separated.
"""

CHASSIS_TYPE_SLED = 'sled chassis type'
"""Sled

An enclosed or semi-enclosed, system chassis which must be plugged into a
multi-system chassis to function normally similar to a blade type chassis.
"""

CHASSIS_TYPE_SHELF = 'shelf chassis type'
"""Shelf

An enclosed or semi-enclosed, typically horizontally-oriented, system chassis
which must be plugged into a multi-system chassis to function
normally.
"""

CHASSIS_TYPE_DRAWER = 'drawer chassis type'
"""Drawer

An enclosed or semi-enclosed, typically horizontally-oriented, system
chassis which may be slid into a multi-system chassis.
"""

CHASSIS_TYPE_MODULE = 'module chassis type'
"""Module

A small, typically removable, chassis or card which contains devices for
a particular subsystem or function.
"""

CHASSIS_TYPE_COMPONENT = 'component chassis type'
"""Component

A small chassis, card, or device which contains devices for a particular
subsystem or function.
"""

CHASSIS_TYPE_IP_BASED_DRIVE = 'IP based drive chassis type'
"""A chassis in a drive form factor with IP-based network connections"""

CHASSIS_TYPE_RACK_GROUP = 'rack group chassis type'
"""A group of racks which form a single entity or share infrastructure"""

CHASSIS_TYPE_STORAGE_ENCLOSURE = 'storage enclosure chassis type'
"""A chassis which encloses storage"""

CHASSIS_TYPE_OTHER = 'other chassis type'
"""A chassis that does not fit any of these definitions"""

# Chassis IntrusionSensor constants

CHASSIS_INTRUSION_SENSOR_NORMAL = 'normal chassis intrusion sensor'
"""No abnormal physical security conditions are detected at this time"""

CHASSIS_INTRUSION_SENSOR_HARDWARE_INTRUSION = 'hardware intrusion chassis ' \
                                              'intrusion sensor'
"""HardwareIntrusion

A door, lock, or other mechanism protecting the internal system hardware from
being accessed is detected as being in an insecure state.
"""

CHASSIS_INTRUSION_SENSOR_TAMPERING_DETECTED = 'tampering detected chassis ' \
                                              'intrusion sensor'
"""Physical tampering of the monitored entity is detected"""

# Chassis IntrusionSensorReArm constants

CHASSIS_INTRUSION_SENSOR_RE_ARM_MANUAL = 'manual re arm chassis intrusion ' \
                                         'sensor'
"""This sensor would be restored to the Normal state by a manual re-arm"""

CHASSIS_INTRUSION_SENSOR_RE_ARM_AUTOMATIC = 'automatic re arm chassis ' \
                                            'intrusion sensor'
"""Automatic

This sensor would be restored to the Normal state automatically as no abnormal
physical security conditions are detected.
"""



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sushy.resources import constants as res_cons
from sushy import utils


STATE_VALUE_MAP = {
    'Enabled': res_cons.STATE_ENABLED,
    'Disabled': res_cons.STATE_DISABLED,
    'Absent': res_cons.STATE_ABSENT,
}

STATE_VALUE_MAP_REV = (
    utils.revert_dictionary(STATE_VALUE_MAP))

HEALTH_VALUE_MAP = {
    'OK': res_cons.HEALTH_OK,
    'Warning': res_cons.HEALTH_WARNING,
    'Critical': res_cons.HEALTH_CRITICAL
}

HEALTH_VALUE_MAP_REV = (
    utils.revert_dictionary(HEALTH_VALUE_MAP))

PARAMTYPE_VALUE_MAP = {
    'string': res_cons.PARAMTYPE_STRING,
    'number': res_cons.PARAMTYPE_NUMBER
}

SEVERITY_VALUE_MAP = {
    'OK': res_cons.SEVERITY_OK,
    'Warning': res_cons.SEVERITY_WARNING,
    'Critical': res_cons.SEVERITY_CRITICAL
}

INDICATOR_LED_VALUE_MAP = {
    'Lit': res_cons.INDICATOR_LED_LIT,
    'Blinking': res_cons.INDICATOR_LED_BLINKING,
    'Off': res_cons.INDICATOR_LED_OFF,
    'Unknown': res_cons.INDICATOR_LED_UNKNOWN,
}

INDICATOR_LED_VALUE_MAP_REV = utils.revert_dictionary(INDICATOR_LED_VALUE_MAP)

POWER_STATE_VALUE_MAP = {
    'On': res_cons.POWER_STATE_ON,
    'Off': res_cons.POWER_STATE_OFF,
    'PoweringOn': res_cons.POWER_STATE_POWERING_ON,
    'PoweringOff': res_cons.POWER_STATE_POWERING_OFF,
}

POWER_STATE_MAP_REV = utils.revert_dictionary(POWER_STATE_VALUE_MAP)

RESET_TYPE_VALUE_MAP = {
    'On': res_cons.RESET_TYPE_ON,
    'ForceOff': res_cons.RESET_TYPE_FORCE_OFF,
    'GracefulShutdown': res_cons.RESET_TYPE_GRACEFUL_SHUTDOWN,
    'GracefulRestart': res_cons.RESET_TYPE_GRACEFUL_RESTART,
    'ForceRestart': res_cons.RESET_TYPE_FORCE_RESTART,
    'Nmi': res_cons.RESET_TYPE_NMI,
    'ForceOn': res_cons.RESET_TYPE_FORCE_ON,
    'PushPowerButton': res_cons.RESET_TYPE_PUSH_POWER_BUTTON,
    'PowerCycle': res_cons.RESET_TYPE_POWER_CYCLE,
}

RESET_TYPE_VALUE_MAP_REV = utils.revert_dictionary(RESET_TYPE_VALUE_MAP)

PROTOCOL_TYPE_VALUE_MAP = {
    'AHCI': res_cons.PROTOCOL_TYPE_AHCI,
    'FC': res_cons.PROTOCOL_TYPE_FC,
    'FCP': res_cons.PROTOCOL_TYPE_FCP,
    'FCoE': res_cons.PROTOCOL_TYPE_FCoE,
    'FICON': res_cons.PROTOCOL_TYPE_FICON,
    'FTP': res_cons.PROTOCOL_TYPE_FTP,
    'HTTP': res_cons.PROTOCOL_TYPE_HTTP,
    'HTTPS': res_cons.PROTOCOL_TYPE_HTTPS,
    'I2C': res_cons.PROTOCOL_TYPE_I2C,
    'NFSv3': res_cons.PROTOCOL_TYPE_NFSv3,
    'NFSv4': res_cons.PROTOCOL_TYPE_NFSv4,
    'NVMe': res_cons.PROTOCOL_TYPE_NVMe,
    'NVMeOverFabrics': res_cons.PROTOCOL_TYPE_NVMeOverFabrics,
    'OEM': res_cons.PROTOCOL_TYPE_OEM,
    'PCIe': res_cons.PROTOCOL_TYPE_PCIe,
    'RoCE': res_cons.PROTOCOL_TYPE_RoCE,
    'RoCEv2': res_cons.PROTOCOL_TYPE_RoCEv2,
    'SAS': res_cons.PROTOCOL_TYPE_SAS,
    'SATA': res_cons.PROTOCOL_TYPE_SATA,
    'SFTP': res_cons.PROTOCOL_TYPE_SFTP,
    'SMB': res_cons.PROTOCOL_TYPE_SMB,
    'UHCI': res_cons.PROTOCOL_TYPE_UHCI,
    'USB': res_cons.PROTOCOL_TYPE_USB,
    'iSCSI': res_cons.PROTOCOL_TYPE_iSCSI,
    'iWARP': res_cons.PROTOCOL_TYPE_iWARP,
}

DUR_NAME_FORMAT_VALUE_MAP = {
    'EUI': res_cons.DURABLE_NAME_FORMAT_EUI,
    'FC_WWN': res_cons.DURABLE_NAME_FORMAT_FC_WWN,
    'NAA': res_cons.DURABLE_NAME_FORMAT_NAA,
    'NQN': res_cons.DURABLE_NAME_FORMAT_NQN,
    'NSID': res_cons.DURABLE_NAME_FORMAT_NSID,
    'UUID': res_cons.DURABLE_NAME_FORMAT_UUID,
    'iQN': res_cons.DURABLE_NAME_FORMAT_iQN,
}

APPLY_TIME_VALUE_MAP = {
    'Immediate': res_cons.APPLY_TIME_IMMEDIATE,
    'OnReset': res_cons.APPLY_TIME_ON_RESET,
    'AtMaintenanceWindowStart':
        res_cons.APPLY_TIME_MAINT_START,
    'InMaintenanceWindowOnReset':
        res_cons.APPLY_TIME_MAINT_RESET,
}

APPLY_TIME_VALUE_MAP_REV = utils.revert_dictionary(APPLY_TIME_VALUE_MAP)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values come from the Redfish UpdateService json-schema.
# https://redfish.dmtf.org/schemas/v1/TaskService.v1_1_5.json#/definitions/OverWritePolicy

# Overwrite Policy constants

OVERWRITE_POLICY_OLDEST = 'oldest completed'
OVERWRITE_POLICY_MANUAL = 'manual only'



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock

from sushy import exceptions
from sushy.resources.sessionservice import session
from sushy.tests.unit import base


class SessionTestCase(base.TestCase):

    def setUp(self):
        super(SessionTestCase, self).setUp()
        self.conn = mock.Mock()
        self.auth = mock.Mock()
        with open('sushy/tests/unit/json_samples/session.json') as f:
            self.json_doc = json.load(f)
            self.conn.get.return_value.json.return_value = self.json_doc
            self.auth._session_key = 'fake_x_auth_token'
            self.auth._session_uri = self.json_doc['@odata.id']
            self.conn._auth = self.auth

        self.sess_inst = session.Session(
            self.conn, '/redfish/v1/SessionService/Sessions/1234567890ABCDEF',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.sess_inst._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.sess_inst.redfish_version)
        self.assertEqual('1234567890ABCDEF', self.sess_inst.identity)
        self.assertEqual('User Session', self.sess_inst.name)
        exp_path = '/redfish/v1/SessionService/Sessions/1234567890ABCDEF'
        self.assertEqual(exp_path, self.sess_inst.path)

    def test__parse_attributes_missing_identity(self):
        self.sess_inst.json.pop('Id')
        self.assertRaisesRegex(
            exceptions.MissingAttributeError, 'attribute Id',
            self.sess_inst._parse_attributes, self.json_doc)

    def test_session_close(self):
        session_key = self.sess_inst._conn._auth._session_key
        session_uri = self.sess_inst._conn._auth._session_uri
        self.assertEqual(session_key, 'fake_x_auth_token')
        self.assertEqual(session_uri, self.sess_inst.path)
        self.sess_inst.delete()
        self.sess_inst._conn.delete.assert_called_with(session_uri)


class SessionCollectionTestCase(base.TestCase):

    def setUp(self):
        super(SessionCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'session_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.sess_col = session.SessionCollection(
            self.conn, '/redfish/v1/SessionService/Sessions',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        path = '/redfish/v1/SessionService/Sessions/104f9d68f58abb85'
        self.sess_col._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.sess_col.redfish_version)
        self.assertEqual('Session Collection', self.sess_col.name)
        self.assertEqual((path,), self.sess_col.members_identities)

    @mock.patch.object(session, 'Session', autospec=True)
    def test_get_member(self, mock_session):
        path = '/redfish/v1/SessionService/Sessions/104f9d68f58abb85'
        self.sess_col.get_member(path)
        mock_session.assert_called_once_with(
            self.sess_col._conn, path,
            self.sess_col.redfish_version, None)

    @mock.patch.object(session, 'Session', autospec=True)
    def test_get_members(self, mock_session):
        path = '/redfish/v1/SessionService/Sessions/104f9d68f58abb85'
        members = self.sess_col.get_members()
        mock_session.assert_called_once_with(
            self.sess_col._conn, path,
            self.sess_col.redfish_version, None)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from http import client as http_client
import json
from unittest import mock

from dateutil import parser

from sushy import exceptions
from sushy.resources import constants as res_cons
from sushy.resources.registry import message_registry
from sushy.resources import settings
from sushy.resources.system import bios
from sushy.tests.unit import base


class BiosTestCase(base.TestCase):

    def setUp(self):
        super(BiosTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/bios.json') as f:
            self.bios_json = json.load(f)
        with open('sushy/tests/unit/json_samples/bios_settings.json') as f:
            self.bios_settings_json = json.load(f)

        self.conn.get.return_value.json.side_effect = [
            self.bios_json,
            self.bios_settings_json,
            self.bios_settings_json]

        conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/message_registry.json') as f:
            conn.get.return_value.json.return_value = json.load(f)
            registry = message_registry.MessageRegistry(
                conn, '/redfish/v1/Registries/Test',
                redfish_version='1.0.2')

        self.sys_bios = bios.Bios(
            self.conn, '/redfish/v1/Systems/437XR1138R2/BIOS',
            registries={'Test.1.0': registry},
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.sys_bios._parse_attributes(self.bios_json)
        self.assertEqual('1.0.2', self.sys_bios.redfish_version)
        self.assertEqual('BIOS', self.sys_bios.identity)
        self.assertEqual('BIOS Configuration Current Settings',
                         self.sys_bios.name)
        self.assertIsNone(self.sys_bios.description)
        self.assertEqual('BiosAttributeRegistryP89.v1_0_0',
                         self.sys_bios._attribute_registry)
        self.assertEqual('', self.sys_bios.attributes['AdminPhone'])
        self.assertEqual('Uefi', self.sys_bios.attributes['BootMode'])
        self.assertEqual(0, self.sys_bios.attributes['ProcCoreDisable'])
        self.assertEqual([res_cons.APPLY_TIME_ON_RESET,
                          res_cons.APPLY_TIME_MAINT_RESET],
                         self.sys_bios.supported_apply_times)
        self.assertEqual(600, self.sys_bios.maintenance_window
                         .maintenance_window_duration_in_seconds)
        self.assertEqual(parser.parse('2020-09-01T04:30:00-06:00'),
                         self.sys_bios.maintenance_window
                         .maintenance_window_start_time)
        # testing here if settings subfield parsed by checking ETag,
        # other settings fields tested in specific settings test
        self.assertEqual('9234ac83b9700123cc32',
                         self.sys_bios._settings._etag)
        self.assertEqual('(404) 555-1212',
                         self.sys_bios.pending_attributes['AdminPhone'])
        self.assertEqual(settings.UPDATE_FAILURE,
                         self.sys_bios.update_status.status)

    def test__parse_attributes_return(self):
        attributes = self.sys_bios._parse_attributes(self.bios_json)

        # Test that various types are returned correctly
        self.assertEqual('BIOS Configuration Current Settings',
                         attributes.get('name'))
        self.assertEqual({'AdminPhone': '',
                          'BootMode': 'Uefi',
                          'EmbeddedSata': 'Raid',
                          'NicBoot1': 'NetworkBoot',
                          'NicBoot2': 'Disabled',
                          'PowerProfile': 'MaxPerf',
                          'ProcCoreDisable': 0,
                          'ProcHyperthreading': 'Enabled',
                          'ProcTurboMode': 'Enabled',
                          'UsbControl': 'UsbEnabled'},
                         attributes.get('attributes'))
        self.assertEqual({'maintenance_window_duration_in_seconds': 600,
                         'maintenance_window_start_time':
                          parser.parse('2020-09-01T04:30:00-06:00')},
                         attributes.get('maintenance_window'))

    def test_set_attribute(self):
        self.sys_bios.set_attribute('ProcTurboMode', 'Disabled')
        self.sys_bios._conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Settings',
            data={'Attributes': {'ProcTurboMode': 'Disabled'}})

    def test_set_attribute_apply_time(self):
        self.sys_bios.set_attribute('ProcTurboMode', 'Disabled',
                                    res_cons.APPLY_TIME_ON_RESET)
        self.sys_bios._conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Settings',
            data={'Attributes': {'ProcTurboMode': 'Disabled'},
                  '@Redfish.SettingsApplyTime': {
                      '@odata.type': '#Settings.v1_0_0.PreferredApplyTime',
                      'ApplyTime': 'OnReset'}})

    def test_set_attribute_apply_time_with_maintenance_window(self):
        self.sys_bios.set_attribute('ProcTurboMode', 'Disabled',
                                    res_cons.APPLY_TIME_MAINT_RESET,
                                    datetime.datetime(2020, 9, 1, 4, 30, 0),
                                    600)
        self.sys_bios._conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Settings',
            data={'Attributes': {'ProcTurboMode': 'Disabled'},
                  '@Redfish.SettingsApplyTime': {
                      '@odata.type': '#Settings.v1_0_0.PreferredApplyTime',
                      'ApplyTime': 'InMaintenanceWindowOnReset',
                      'MaintenanceWindowStartTime': '2020-09-01T04:30:00',
                      'MaintenanceWindowDurationInSeconds': 600}})

    def test_set_attribute_on_refresh(self):
        self.conn.get.reset_mock()
        # make it to instantiate pending attributes
        self.sys_bios.pending_attributes
        self.assertTrue(self.conn.get.called)

        self.conn.get.reset_mock()

        self.sys_bios.pending_attributes
        self.assertFalse(self.conn.get.called)

        self.sys_bios.set_attribute('ProcTurboMode', 'Disabled')
        # make it to refresh pending attributes on next retrieval
        self.sys_bios.pending_attributes
        self.assertTrue(self.conn.get.called)

    def test_set_attributes(self):
        self.sys_bios.set_attributes({'ProcTurboMode': 'Disabled',
                                      'UsbControl': 'UsbDisabled'})
        self.sys_bios._conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Settings',
            data={'Attributes': {'ProcTurboMode': 'Disabled',
                                 'UsbControl': 'UsbDisabled'}})

    def test_set_attributes_apply_time(self):
        self.sys_bios.set_attributes({'ProcTurboMode': 'Disabled',
                                      'UsbControl': 'UsbDisabled'},
                                     res_cons.APPLY_TIME_IMMEDIATE)
        self.sys_bios._conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Settings',
            data={'Attributes': {'ProcTurboMode': 'Disabled',
                                 'UsbControl': 'UsbDisabled'},
                  '@Redfish.SettingsApplyTime': {
                      '@odata.type': '#Settings.v1_0_0.PreferredApplyTime',
                      'ApplyTime': 'Immediate'}})

    def test_set_attributes_apply_time_with_maintenance_window(self):
        self.sys_bios.set_attributes({'ProcTurboMode': 'Disabled',
                                      'UsbControl': 'UsbDisabled'},
                                     res_cons.APPLY_TIME_MAINT_START,
                                     datetime.datetime(2020, 9, 1, 4, 30, 0),
                                     600)
        self.sys_bios._conn.patch.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Settings',
            data={'Attributes': {'ProcTurboMode': 'Disabled',
                                 'UsbControl': 'UsbDisabled'},
                  '@Redfish.SettingsApplyTime': {
                      '@odata.type': '#Settings.v1_0_0.PreferredApplyTime',
                      'ApplyTime': 'AtMaintenanceWindowStart',
                      'MaintenanceWindowStartTime': '2020-09-01T04:30:00',
                      'MaintenanceWindowDurationInSeconds': 600}})

    def test_set_attributes_apply_time_missing(self):
        self.assertRaises(ValueError,
                          self.sys_bios.set_attributes,
                          {'ProcTurboMode': 'Disabled',
                           'UsbControl': 'UsbDisabled'},
                          maint_window_start_time=datetime.datetime.now(),
                          maint_window_duration=600)

    def test_set_attributes_maint_window_start_time_missing(self):
        self.assertRaises(ValueError,
                          self.sys_bios.set_attributes,
                          {'ProcTurboMode': 'Disabled',
                           'UsbControl': 'UsbDisabled'},
                          res_cons.APPLY_TIME_MAINT_START,
                          maint_window_duration=600)

    def test_set_attributes_maint_window_duration_missing(self):
        self.assertRaises(ValueError,
                          self.sys_bios.set_attributes,
                          {'ProcTurboMode': 'Disabled',
                           'UsbControl': 'UsbDisabled'},
                          res_cons.APPLY_TIME_MAINT_START,
                          datetime.datetime.now())

    def test_set_attributes_on_refresh(self):
        self.conn.get.reset_mock()
        # make it to instantiate pending attributes
        self.sys_bios.pending_attributes
        self.assertTrue(self.conn.get.called)

        self.conn.get.reset_mock()

        self.sys_bios.pending_attributes
        self.assertFalse(self.conn.get.called)

        self.sys_bios.set_attributes({'ProcTurboMode': 'Disabled',
                                      'UsbControl': 'UsbDisabled'})
        # make it to refresh pending attributes on next retrieval
        self.sys_bios.pending_attributes
        self.assertTrue(self.conn.get.called)

    def test_apply_time_settings(self):
        self.conn.get.reset_mock()
        apply_time_settings = self.sys_bios.apply_time_settings
        self.assertIsNotNone(apply_time_settings)
        self.assertEqual('OnReset', apply_time_settings.apply_time)
        self.assertEqual(['OnReset', 'Immediate', 'AtMaintenanceWindowStart',
                         'InMaintenanceWindowOnReset'],
                         apply_time_settings.apply_time_allowable_values)
        self.assertEqual(parser.parse('2017-05-03T23:12:37-05:00'),
                         apply_time_settings.maintenance_window_start_time)
        self.assertEqual(600, apply_time_settings.
                         maintenance_window_duration_in_seconds)

    def test__get_reset_bios_action_element(self):
        value = self.sys_bios._get_reset_bios_action_element()
        self.assertEqual('/redfish/v1/Systems/437XR1138R2/BIOS/Actions/'
                         'Bios.ResetBios',
                         value.target_uri)

    def test_reset_bios_missing_action(self):
        self.sys_bios._actions.reset_bios = None
        self.assertRaisesRegex(
            exceptions.MissingActionError, '#Bios.ResetBios',
            self.sys_bios.reset_bios)

    def test__parse_attributes_missing_reset_bios_target(self):
        self.sys_bios.json['Actions']['#Bios.ResetBios'].pop(
            'target')
        self.assertRaisesRegex(
            exceptions.MissingAttributeError,
            'attribute Actions/#Bios.ResetBios/target',
            self.sys_bios._parse_attributes, self.bios_json)

    def test_reset_bios(self):
        self.sys_bios.reset_bios()
        self.sys_bios._conn.post.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Actions/Bios.ResetBios')

    def test_reset_bios_handle_http_error_415(self):

        target_uri = (
            '/redfish/v1/Systems/437XR1138R2/BIOS/Actions/Bios.ResetBios')
        self.conn.post.side_effect = [exceptions.HTTPError(
            method='POST', url=target_uri, response=mock.MagicMock(
                status_code=http_client.UNSUPPORTED_MEDIA_TYPE)), '200']
        post_calls = [
            mock.call(target_uri), mock.call(target_uri, data={})]
        self.sys_bios.reset_bios()
        self.sys_bios._conn.post.assert_has_calls(post_calls)

    def test_reset_bios_handle_http_error_400(self):

        target_uri = (
            '/redfish/v1/Systems/437XR1138R2/BIOS/Actions/Bios.ResetBios')
        self.conn.post.side_effect = [exceptions.HTTPError(
            method='POST', url=target_uri, response=mock.MagicMock(
                status_code=http_client.BAD_REQUEST)), '200']
        post_calls = [
            mock.call(target_uri), mock.call(target_uri, data={})]
        self.sys_bios.reset_bios()
        self.sys_bios._conn.post.assert_has_calls(post_calls)

    def test_reset_bios_handle_http_error_405(self):

        target_uri = (
            '/redfish/v1/Systems/437XR1138R2/BIOS/Actions/Bios.ResetBios')
        self.conn.post.side_effect = exceptions.HTTPError(
            method='POST', url=target_uri, response=mock.MagicMock(
                status_code=http_client.METHOD_NOT_ALLOWED))
        self.assertRaises(
            exceptions.HTTPError,
            self.sys_bios.reset_bios)
        self.sys_bios._conn.post.assert_called_once_with(target_uri)

    def test__get_change_password_element(self):
        value = self.sys_bios._get_change_password_element()
        self.assertEqual("/redfish/v1/Systems/437XR1138R2/BIOS/Actions/"
                         "Bios.ChangePassword",
                         value.target_uri)

    def test_change_password_missing_action(self):
        self.sys_bios._actions.change_password = None
        self.assertRaisesRegex(
            exceptions.MissingActionError, '#Bios.ChangePassword',
            self.sys_bios.change_password, 'newpassword',
                                           'oldpassword',
                                           'adminpassword')

    def test__parse_attributes_missing_change_password_target(self):
        self.sys_bios.json['Actions']['#Bios.ChangePassword'].pop(
            'target')
        self.assertRaisesRegex(
            exceptions.MissingAttributeError,
            'attribute Actions/#Bios.ChangePassword/target',
            self.sys_bios._parse_attributes, self.bios_json)

    def test_change_password(self):
        self.sys_bios.change_password('newpassword',
                                      'oldpassword',
                                      'adminpassword')
        self.sys_bios._conn.post.assert_called_once_with(
            '/redfish/v1/Systems/437XR1138R2/BIOS/Actions/Bios.ChangePassword',
            data={'OldPassword': 'oldpassword',
                  'NewPassword': 'newpassword',
                  'PasswordName': 'adminpassword'})



# Copyright (c) 2020 Dell, Inc. or its subsidiaries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from sushy.resources import constants as res_cons
from sushy.resources.taskservice import constants as ts_cons
from sushy import utils


TASK_STATE_VALUE_MAP = {
    'New': res_cons.TASK_STATE_NEW,
    'Starting': res_cons.TASK_STATE_STARTING,
    'Running': res_cons.TASK_STATE_RUNNING,
    'Suspended': res_cons.TASK_STATE_SUSPENDED,
    'Interrupted': res_cons.TASK_STATE_INTERRUPTED,
    'Pending': res_cons.TASK_STATE_PENDING,
    'Stopping': res_cons.TASK_STATE_STOPPING,
    'Completed': res_cons.TASK_STATE_COMPLETED,
    'Killed': res_cons.TASK_STATE_KILLED,
    'Exception': res_cons.TASK_STATE_EXCEPTION,
    'Service': res_cons.TASK_STATE_SERVICE,
    'Cancelling': res_cons.TASK_STATE_CANCELLING,
    'Cancelled': res_cons.TASK_STATE_CANCELLED
}

OVERWRITE_POLICY_VALUE_MAP = {
    'Oldest': ts_cons.OVERWRITE_POLICY_OLDEST,
    'Manual': ts_cons.OVERWRITE_POLICY_MANUAL,
}

OVERWRITE_POLICY_VALUE_MAP_REV = (
    utils.revert_dictionary(OVERWRITE_POLICY_VALUE_MAP))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/ResourceBlock.v1_1_0.json

import logging

from sushy.resources import base
from sushy.resources import common
from sushy.resources.compositionservice import mappings as res_maps

LOG = logging.getLogger(__name__)


class CompositionStatusField(base.CompositeField):

    composition_state = base.MappedField(
        'CompositionState',
        res_maps.COMPOSITION_STATE_VALUE_MAP,
        required=True)
    """Inform the client, state of the resource block"""

    max_compositions = base.Field('MaxCompositions')
    """The maximum number of compositions"""

    number_of_compositions = base.Field('NumberOfCompositions')
    """The number of compositions"""

    reserved_state = base.Field('Reserved')
    """Inform the resource block has been identified by a client"""

    sharing_capable = base.Field('SharingCapable')
    """Indicates if this Resource Block is capable of participating in
    multiple compositions simultaneously"""

    sharing_enabled = base.Field('SharingEnabled')
    """Indicates if this Resource Block is allowed to participate in
    multiple compositions simultaneously"""


class ResourceBlock(base.ResourceBase):

    composition_status = CompositionStatusField(
        'CompositionStatus',
        required=True)
    """The composition state of resource block"""

    description = base.Field('Description')
    """The resource block description"""

    identity = base.Field('Id', required=True)
    """The resource block identity string"""

    name = base.Field('Name', required=True)
    """The resource block name"""

    resource_block_type = base.MappedField(
        'ResourceBlockType',
        res_maps.RESOURCE_BLOCK_TYPE_VALUE_MAP,
        required=True)
    """The type of resource block"""

    status = common.StatusField('Status')
    """The status of resource block"""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a ResourceBlock

        :param connector: A Connector instance
        :param identity: The identity of the ResourceBlock resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        """
        super(ResourceBlock, self).__init__(
            connector, identity, redfish_version, registries)


class ResourceBlockCollection(base.ResourceCollectionBase):

    name = base.Field('Name')
    """The resource block collection name"""

    description = base.Field('Description')
    """The resource block collection description"""

    @property
    def _resource_type(self):
        return ResourceBlock

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a ResourceBlockCollection

        :param connector: A Connector instance
        :param identity: A identity of the ResourceBlock resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        """
        super(ResourceBlockCollection, self).__init__(
            connector, identity, redfish_version, registries)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/EthernetInterface.v1_4_0.json

import logging

from sushy.resources import base
from sushy.resources import common
from sushy.resources import constants as res_cons
from sushy import utils

LOG = logging.getLogger(__name__)


class EthernetInterface(base.ResourceBase):
    """This class adds the EthernetInterface resource"""

    identity = base.Field('Id', required=True)
    """The Ethernet Interface identity string"""

    name = base.Field('Name')
    """The name of the resource or array element"""

    description = base.Field('Description')
    """Description"""

    permanent_mac_address = base.Field('PermanentMACAddress')
    """This is the permanent MAC address assigned to this interface (port) """

    mac_address = base.Field('MACAddress')
    """This is the currently configured MAC address of the interface."""

    speed_mbps = base.Field('SpeedMbps')
    """This is the current speed in Mbps of this interface."""

    status = common.StatusField("Status")
    """Describes the status and health of this interface."""


class EthernetInterfaceCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return EthernetInterface

    @property
    @utils.cache_it
    def summary(self):
        """Summary of MAC addresses and interfaces state

        This filters the MACs whose health is OK,
        which means the MACs in both 'Enabled' and 'Disabled' States
        are returned.

        :returns: dictionary in the format
            {'aa:bb:cc:dd:ee:ff': sushy.STATE_ENABLED,
            'aa:bb:aa:aa:aa:aa': sushy.STATE_DISABLED}
        """
        mac_dict = {}
        for eth in self.get_members():
            if eth.mac_address is not None and eth.status is not None:
                if eth.status.health == res_cons.HEALTH_OK:
                    mac_dict[eth.mac_address] = eth.status.state
        return mac_dict



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock


from sushy.resources import constants as res_cons
from sushy.resources.system import simple_storage
from sushy.tests.unit import base


class SimpleStorageTestCase(base.TestCase):

    def setUp(self):
        super(SimpleStorageTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.simpl_stor = simple_storage.SimpleStorage(
            self.conn, '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.simpl_stor._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.simpl_stor.redfish_version)
        self.assertEqual('1', self.simpl_stor.identity)
        self.assertEqual('Simple Storage Controller', self.simpl_stor.name)
        self.assertEqual(8000000000000,
                         self.simpl_stor.devices[0].capacity_bytes)
        self.assertEqual(4000000000000,
                         self.simpl_stor.devices[1].capacity_bytes)
        self.assertEqual(res_cons.STATE_ENABLED,
                         self.simpl_stor.devices[0].status.state)
        self.assertEqual(res_cons.STATE_ABSENT,
                         self.simpl_stor.devices[2].status.state)
        self.assertEqual(res_cons.HEALTH_OK,
                         self.simpl_stor.devices[0].status.health)


class SimpleStorageCollectionTestCase(base.TestCase):

    def setUp(self):
        super(SimpleStorageCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'simple_storage_collection.json') as f:

            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.simpl_stor_col = simple_storage.SimpleStorageCollection(
            self.conn, '/redfish/v1/Systems/437XR1138R2/SimpleStorage',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.simpl_stor_col._parse_attributes(self.json_doc)
        self.assertEqual((
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',),
            self.simpl_stor_col.members_identities)

    @mock.patch.object(simple_storage, 'SimpleStorage', autospec=True)
    def test_get_member(self, SimpleStorage_mock):
        self.simpl_stor_col.get_member(
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1')
        SimpleStorage_mock.assert_called_once_with(
            self.simpl_stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',
            self.simpl_stor_col.redfish_version, None)

    @mock.patch.object(simple_storage, 'SimpleStorage', autospec=True)
    def test_get_members(self, SimpleStorage_mock):
        members = self.simpl_stor_col.get_members()
        SimpleStorage_mock.assert_called_once_with(
            self.simpl_stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/SimpleStorage/1',
            self.simpl_stor_col.redfish_version, None)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))

    def test_disks_sizes_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.assertEqual([4000000000000, 8000000000000],
                         self.simpl_stor_col.disks_sizes_bytes)

    def test_disks_sizes_bytes_capacity_bytes_none(self):
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            json_doc = json.load(f)

        json_doc['Devices'][0]['CapacityBytes'] = None
        self.conn.get.return_value.json.return_value = json_doc

        self.assertEqual([4000000000000],
                         self.simpl_stor_col.disks_sizes_bytes)

    def test_max_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.assertEqual(8000000000000, self.simpl_stor_col.max_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(8000000000000, self.simpl_stor_col.max_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_max_size_bytes_after_refresh(self):
        self.simpl_stor_col.refresh()
        self.conn.get.return_value.json.reset_mock()

        with open('sushy/tests/unit/json_samples/'
                  'simple_storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.assertEqual(8000000000000, self.simpl_stor_col.max_size_bytes)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Values come from the Redfish UpdateService json-schema.
# https://redfish.dmtf.org/schemas/UpdateService.v1_2_2.json#/definitions/TransferProtocolType

from sushy.resources import constants as res_cons

# Transfer Protocol Type constants

UPDATE_PROTOCOL_CIFS = res_cons.PROTOCOL_TYPE_CIFS
UPDATE_PROTOCOL_FTP = res_cons.PROTOCOL_TYPE_FTP
UPDATE_PROTOCOL_SFTP = res_cons.PROTOCOL_TYPE_SFTP
UPDATE_PROTOCOL_HTTP = res_cons.PROTOCOL_TYPE_HTTP
UPDATE_PROTOCOL_HTTPS = res_cons.PROTOCOL_TYPE_HTTPS
UPDATE_PROTOCOL_SCP = res_cons.PROTOCOL_TYPE_SCP
UPDATE_PROTOCOL_TFTP = res_cons.PROTOCOL_TYPE_TFTP
UPDATE_PROTOCOL_OEM = res_cons.PROTOCOL_TYPE_OEM
UPDATE_PROTOCOL_NFS = res_cons.PROTOCOL_TYPE_NFS



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/SessionService.v1_1_3.json

import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources.sessionservice import session
from sushy import utils

LOG = logging.getLogger(__name__)


class SessionService(base.ResourceBase):

    description = base.Field('Description')
    """The session service description"""

    identity = base.Field('Id', required=True)
    """The session service identify string"""

    name = base.Field('Name', required=True)
    """The session service name"""

    service_enabled = base.Field('ServiceEnabled')
    """Tells us if session service is enabled"""

    session_timeout = base.Field('SessionTimeout')
    """The session service timeout"""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a SessionService

        Warning: This class should only be invoked with a connector which
        has already established authentication.

        :param connector: A Connector instance
        :param identity: The identity of the SessionService resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        """
        # Populating the base resource so session interactions can
        # occur based on the contents of it.
        super(SessionService, self).__init__(
            connector, identity, redfish_version=redfish_version,
            registries=registries)

    def _get_sessions_collection_path(self):
        """Helper function to find the SessionCollections path"""
        sessions_col = self.json.get('Sessions')
        if not sessions_col:
            raise exceptions.MissingAttributeError(
                attribute='Sessions', resource=self._path)
        return sessions_col.get('@odata.id')

    @property
    @utils.cache_it
    def sessions(self):
        """Property to provide reference to the `SessionCollection` instance

        It is calculated once when the first time it is queried. On refresh,
        this property gets reset.
        """
        return session.SessionCollection(
            self._conn, self._get_sessions_collection_path(),
            self.redfish_version, self.registries)

    def close_session(self, session_uri):
        """This function is for closing a session based on its id.

        :raises: ServerSideError
        """
        self._conn.delete(session_uri)

    def create_session(self, username, password, target_uri=None):
        """This function will try to create a session.

        Create a session and return the associated key and URI.

        :param username: the username of the user requesting a new session
        :param password: the password associated to the user requesting
            a new session
        :param target_uri: the "Sessions" uri, usually in the form:
            '/redfish/v1/SessionService/Sessions'
        :returns: A session key and uri in the form of a tuple
        :raises: MissingXAuthToken
        :raises: ConnectionError
        :raises: AccessError
        :raises: HTTPError
        """
        if not target_uri:
            try:
                target_uri = self._get_sessions_collection_path()
            except Exception:
                # Defaulting to /Sessions
                target_uri = self.path + '/Sessions'

        data = {'UserName': username, 'Password': password}
        LOG.debug("Requesting new session from %s.", target_uri)
        rsp = self._conn.post(target_uri, data=data)
        session_key = rsp.headers.get('X-Auth-Token')
        if session_key is None:
            raise exceptions.MissingXAuthToken(
                method='POST', url=target_uri, response=rsp)

        session_uri = rsp.headers.get('Location')
        if session_uri is None:
            LOG.warning("Received X-Auth-Token but NO session uri.")

        return session_key, session_uri



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/UpdateService.v1_2_2.json

import logging

from sushy import exceptions
from sushy.resources import base
from sushy.resources import common
from sushy.resources.updateservice import constants as up_cons
from sushy.resources.updateservice import mappings as up_maps
from sushy.resources.updateservice import softwareinventory
from sushy import taskmonitor
from sushy import utils

LOG = logging.getLogger(__name__)


class ActionsField(base.CompositeField):

    simple_update = common.ActionField('#UpdateService.SimpleUpdate')


class UpdateService(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The update service identity"""

    http_push_uri = base.Field('HttpPushUri')
    """The URI used to perform an HTTP or HTTPS push update to the Update
    Service"""

    http_push_uri_targets = base.Field('HttpPushUriTargets')
    """The array of URIs indicating the target for applying the""" + \
        """update image"""

    http_push_uri_targets_busy = base.Field('HttpPushUriTargetsBusy')
    """This represents if the HttpPushUriTargets property is reserved""" + \
        """by anyclient"""

    name = base.Field('Name', required=True)
    """The update service name"""

    service_enabled = base.Field('ServiceEnabled')
    """The status of whether this service is enabled"""

    status = common.StatusField('Status')
    """The status of the update service"""

    _actions = ActionsField('Actions', required=True)

    _firmware_inventory_path = base.Field(['FirmwareInventory', '@odata.id'])

    _software_inventory_path = base.Field(['SoftwareInventory', '@odata.id'])

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None):
        """A class representing a UpdateService

        :param connector: A Connector instance
        :param identity: The identity of the UpdateService resource
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of given version
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        """
        super(UpdateService, self).__init__(
            connector, identity, redfish_version, registries)

    def _get_simple_update_element(self):
        simple_update_action = self._actions.simple_update
        if not simple_update_action:
            raise exceptions.MissingAttributeError(
                action='#UpdateService.SimpleUpdate',
                resource=self._path)
        return simple_update_action

    def _get_legacy_transfer_protocols(self):
        """Get the backward-compatible values for transfer protocol.

        :returns: A set of allowed values.
        """
        LOG.warning(
            'Could not figure out the allowed values for the simple '
            'update action for UpdateService %s', self.identity)
        return set(up_maps.TRANSFER_PROTOCOL_TYPE_VALUE_MAP)

    def get_allowed_transfer_protocols(self):
        """Get the allowed values for transfer protocol.

        :returns: A set of allowed values.
        :raises: MissingAttributeError, if Actions/#UpdateService.SimpleUpdate
            attribute not present.
        """
        simple_update_action = self._get_simple_update_element()

        if not getattr(simple_update_action, 'transfer_protocol', None):
            LOG.debug(
                'Server does not constrain allowed transfer protocols for '
                'simple update action of UpdateService %s', self.identity)
            return set(up_maps.TRANSFER_PROTOCOL_TYPE_VALUE_MAP_REV)

        return {simple_update_action.transfer_protocol}

    def simple_update(self, image_uri, targets=None,
                      transfer_protocol=up_cons.UPDATE_PROTOCOL_HTTP):
        """Simple Update is used to update software components.

        :returns: A task monitor.
        """
        valid_transfer_protocols = self.get_allowed_transfer_protocols()

        if transfer_protocol in valid_transfer_protocols:
            transfer_protocol = up_maps.TRANSFER_PROTOCOL_TYPE_VALUE_MAP_REV[
                transfer_protocol]

        else:
            legacy_transfer_protocols = self._get_legacy_transfer_protocols()

            if transfer_protocol not in legacy_transfer_protocols:
                raise exceptions.InvalidParameterValueError(
                    parameter='transfer_protocol', value=transfer_protocol,
                    valid_values=valid_transfer_protocols)

            LOG.warning(
                'Legacy transfer protocol constant %s is being used. '
                'Consider migrating to any of: %s', transfer_protocol,
                ', '.join(up_maps.TRANSFER_PROTOCOL_TYPE_VALUE_MAP_REV))

        target_uri = self._get_simple_update_element().target_uri

        LOG.debug(
            'Updating software component %s via '
            '%s ...', image_uri, target_uri)

        data = {'ImageURI': image_uri, 'TransferProtocol': transfer_protocol}
        if targets:
            data['Targets'] = targets
        rsp = self._conn.post(target_uri, data=data)

        return taskmonitor.TaskMonitor.from_response(
            self._conn, rsp, target_uri, self.redfish_version, self.registries)

    def get_task_monitor(self, task_monitor):
        """Used to retrieve a TaskMonitor.

        Deprecated: Use sushy.Sushy.get_task_monitor
        :returns: A task monitor.
        """
        return taskmonitor.TaskMonitor(
            self._conn,
            task_monitor,
            redfish_version=self.redfish_version,
            registries=self.registries)

    @property
    @utils.cache_it
    def software_inventory(self):
        """Property to reference SoftwareInventory collection instance"""
        if not self._software_inventory_path:
            raise exceptions.MissingAttributeError(
                attribute='SoftwareInventory/@odata.id',
                resource=self._software_inventory_path)

        return softwareinventory.SoftwareInventoryCollection(
            self._conn,
            self._software_inventory_path,
            redfish_version=self.redfish_version,
            registries=self.registries)

    @property
    @utils.cache_it
    def firmware_inventory(self):
        """Property to reference FirmwareInventory collection instance"""
        if not self._firmware_inventory_path:
            raise exceptions.MissingAttributeError(
                attribute='FirmwareInventory/@odata.id',
                resource=self._firmware_inventory_path)

        return softwareinventory.SoftwareInventoryCollection(
            self._conn,
            self._firmware_inventory_path,
            redfish_version=self.redfish_version,
            registries=self.registries)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from dateutil import parser

from sushy.resources import base
from sushy.resources import mappings as res_maps


class IdRefField(base.CompositeField):
    """Reference to the resource odata identity field."""

    resource_uri = base.Field('@odata.id')
    """The unique identifier for a resource"""


class OperationApplyTimeSupportField(base.CompositeField):
    def __init__(self):
        super(OperationApplyTimeSupportField, self).__init__(
            path="@Redfish.OperationApplyTimeSupport")

    maintenance_window_duration_in_seconds = base.Field(
        'MaintenanceWindowDurationInSeconds', adapter=int)
    """The expiry time of maintenance window in seconds"""

    _maintenance_window_resource = IdRefField('MaintenanceWindowResource')
    """The location of the maintenance window settings"""

    maintenance_window_start_time = base.Field(
        'MaintenanceWindowStartTime',
        adapter=parser.parse)
    """The start time of a maintenance window"""

    supported_values = base.Field('SupportedValues', required=True,
                                  adapter=list)
    """The types of apply times that the client is allowed request when
    performing a create, delete, or action operation returned as an unmapped
    list

    Deprecated: Use `mapped_supported_values`.
    """

    mapped_supported_values = base.MappedListField(
        'SupportedValues', res_maps.APPLY_TIME_VALUE_MAP, required=True)
    """The types of apply times that the client is allowed request when
    performing a create, delete, or action operation returned as a mapped
    list"""


class ActionField(base.CompositeField):
    target_uri = base.Field('target', required=True)
    operation_apply_time_support = OperationApplyTimeSupportField()


class ResetActionField(ActionField):
    allowed_values = base.Field('ResetType@Redfish.AllowableValues',
                                adapter=list)


class InitializeActionField(ActionField):
    allowed_values = base.Field('InitializeType@Redfish.AllowableValues',
                                adapter=list)


class StatusField(base.CompositeField):
    """This Field describes the status of a resource and its children.

    This field shall contain any state or health properties of a resource.
    """
    health = base.MappedField('Health', res_maps.HEALTH_VALUE_MAP)
    """Represents health of resource w/o considering its dependent resources"""

    health_rollup = base.MappedField('HealthRollup', res_maps.HEALTH_VALUE_MAP)
    """Represents health state of resource and its dependent resources"""

    state = base.MappedField('State', res_maps.STATE_VALUE_MAP)
    """Indicates the known state of the resource, such as if it is enabled."""


class IdentifiersListField(base.ListField):
    """This type describes any additional identifiers for a resource."""

    durable_name = base.Field('DurableName')
    """This indicates the world wide, persistent name of the resource."""

    durable_name_format = base.MappedField('DurableNameFormat',
                                           res_maps.DUR_NAME_FORMAT_VALUE_MAP)
    """This represents the format of the DurableName property."""



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sushy.resources.system.storage import constants as store_cons
from sushy import utils

VOLUME_INIT_TYPE_MAP = {
    'Fast': store_cons.VOLUME_INIT_TYPE_FAST,
    'Slow': store_cons.VOLUME_INIT_TYPE_SLOW
}

VOLUME_INIT_TYPE_MAP_REV = (
    utils.revert_dictionary(VOLUME_INIT_TYPE_MAP)
)

VOLUME_TYPE_TYPE_MAP = {
    'RawDevice': store_cons.VOLUME_TYPE_RAW_DEVICE,
    'NonRedundant': store_cons.VOLUME_TYPE_NON_REDUNDANT,
    'Mirrored': store_cons.VOLUME_TYPE_MIRRORED,
    'StripedWithParity': store_cons.VOLUME_TYPE_STRIPED_WITH_PARITY,
    'SpannedMirrors': store_cons.VOLUME_TYPE_SPANNED_MIRRORS,
    'SpannedStripesWithParity':
        store_cons.VOLUME_TYPE_SPANNED_STRIPES_WITH_PARITY
}

RAID_TYPE_TYPE_MAP = {
    'RAID0': store_cons.RAID_TYPE_RAID0,
    'RAID1': store_cons.RAID_TYPE_RAID1,
    'RAID3': store_cons.RAID_TYPE_RAID3,
    'RAID4': store_cons.RAID_TYPE_RAID4,
    'RAID5': store_cons.RAID_TYPE_RAID5,
    'RAID6': store_cons.RAID_TYPE_RAID6,
    'RAID10': store_cons.RAID_TYPE_RAID10,
    'RAID01': store_cons.RAID_TYPE_RAID01,
    'RAID6TP': store_cons.RAID_TYPE_RAID6TP,
    'RAID1E': store_cons.RAID_TYPE_RAID1E,
    'RAID50': store_cons.RAID_TYPE_RAID50,
    'RAID60': store_cons.RAID_TYPE_RAID60,
    'RAID00': store_cons.RAID_TYPE_RAID00,
    'RAID10E': store_cons.RAID_TYPE_RAID10E,
    'RAID1Triple': store_cons.RAID_TYPE_RAID1Triple,
    'RAID10Triple': store_cons.RAID_TYPE_RAID10Triple,
}



# Copyright (c) 2020 Dell, Inc. or its subsidiaries
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/Task.v1_4_3.json

from http import client as http_client
import logging

from sushy.resources import base
from sushy.resources import mappings as res_maps
from sushy.resources.registry import message_registry
from sushy.resources.taskservice import mappings as task_maps
from sushy import utils


LOG = logging.getLogger(__name__)


class Task(base.ResourceBase):

    identity = base.Field('Id', required=True)
    """The Task identity"""

    name = base.Field('Name', required=True)
    """The Task name"""

    description = base.Field('Description')
    """The Task description"""

    task_monitor = base.Field('TaskMonitor')
    """An opaque URL that the client can use to monitor an asynchronous
       operation"""

    start_time = base.Field('StartTime')
    """Start time of the Task"""

    end_time = base.Field('EndTime')
    """End time of the Task"""

    percent_complete = base.Field('PercentComplete', adapter=utils.int_or_none)
    """Percentage complete of the Task"""

    task_state = base.MappedField('TaskState', task_maps.TASK_STATE_VALUE_MAP)
    """The Task state"""

    task_status = base.MappedField('TaskStatus', res_maps.HEALTH_VALUE_MAP)
    """The Task status"""

    messages = base.MessageListField("Messages")
    """List of :class:`.MessageListField` with messages from the Task"""

    def __init__(self, connector, identity, redfish_version=None,
                 registries=None, json_doc=None):
        """A class representing a Task

        :param connector: A Connector instance
        :param identity: The identity of the task
        :param redfish_version: The version of RedFish. Used to construct
            the object according to schema of the given version.
        :param registries: Dict of Redfish Message Registry objects to be
            used in any resource that needs registries to parse messages
        :param field_data: the data to use populating the fields
        """
        super(Task, self).__init__(
            connector, identity, redfish_version, registries,
            json_doc=json_doc)

    @property
    def is_processing(self):
        """Indicates if the Task is processing"""
        return self.status_code == http_client.ACCEPTED

    def parse_messages(self):
        """Parses the messages"""
        for m in self.messages:
            message_registry.parse_message(self._registries, m)


class TaskCollection(base.ResourceCollectionBase):

    @property
    def _resource_type(self):
        return Task

    @property
    @utils.cache_it
    def summary(self):
        """Summary of task ids and corresponding state

        :returns: dictionary in the format
            {'jid_123456789': sushy.TASK_STATE_NEW,
            'jid_123454321': sushy.TASK_STATE_RUNNING}
        """
        task_dict = {}
        for task in self.get_members():
            task_dict[task.identity] = task.task_state
        return task_dict



# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock


import sushy
from sushy import exceptions
from sushy.resources import constants as res_cons
from sushy.resources.system import processor
from sushy.tests.unit import base


class ProcessorTestCase(base.TestCase):

    def setUp(self):
        super(ProcessorTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/processor.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.sys_processor = processor.Processor(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Processors/CPU1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.sys_processor._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.sys_processor.redfish_version)
        self.assertEqual('CPU1', self.sys_processor.identity)
        self.assertEqual('CPU 1', self.sys_processor.socket)
        self.assertEqual(
            sushy.PROCESSOR_TYPE_CPU,
            self.sys_processor.processor_type)
        self.assertEqual(sushy.PROCESSOR_ARCH_x86,
                         self.sys_processor.processor_architecture)
        self.assertEqual(
            sushy.PROCESSOR_INSTRUCTIONSET_x86_64,
            self.sys_processor.instruction_set)
        self.assertEqual('Intel(R) Corporation',
                         self.sys_processor.manufacturer)
        self.assertEqual('Multi-Core Intel(R) Xeon(R) processor 7xxx Series',
                         self.sys_processor.model)
        self.assertEqual('0x42',
                         self.sys_processor.processor_id.effective_family)
        self.assertEqual('0x61',
                         self.sys_processor.processor_id.effective_model)
        self.assertEqual('0x34AC34DC8901274A',
                         self.sys_processor.processor_id.
                         identification_registers)
        self.assertEqual('0x429943',
                         self.sys_processor.processor_id.microcode_info)
        self.assertEqual('0x1',
                         self.sys_processor.processor_id.step)
        self.assertEqual('GenuineIntel',
                         self.sys_processor.processor_id.vendor_id)

        self.assertEqual(3700, self.sys_processor.max_speed_mhz)
        self.assertEqual(8, self.sys_processor.total_cores)
        self.assertEqual(16, self.sys_processor.total_threads)
        self.assertEqual(res_cons.STATE_ENABLED,
                         self.sys_processor.status.state)
        self.assertEqual(res_cons.HEALTH_OK, self.sys_processor.status.health)
        self.assertEqual(res_cons.HEALTH_OK,
                         self.sys_processor.status.health_rollup)

    def test_sub_processors(self):
        self.conn.get.return_value.json.reset_mock()
        with open('sushy/tests/unit/json_samples/'
                  'subprocessor_collection.json') as f:
            subproc_col = json.load(f)
        with open('sushy/tests/unit/json_samples/subprocessor.json') as f:
            subproc1 = json.load(f)

        self.conn.get.return_value.json.side_effect = [subproc_col, subproc1]

        sub_processors = self.sys_processor.sub_processors
        self.assertIsInstance(sub_processors,
                              processor.ProcessorCollection)
        self.assertEqual((0, None), sub_processors.summary)
        self.assertEqual('Core1', sub_processors.get_members()[0].identity)

    def test_sub_processors_missing(self):
        self.sys_processor.json.pop('SubProcessors')
        with self.assertRaisesRegex(
                exceptions.MissingAttributeError, 'attribute SubProcessors'):
            self.sys_processor.sub_processors


class ProcessorCollectionTestCase(base.TestCase):

    def setUp(self):
        super(ProcessorCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'processor_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.sys_processor_col = processor.ProcessorCollection(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Processors',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.sys_processor_col._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.sys_processor_col.redfish_version)
        self.assertEqual('Processors Collection', self.sys_processor_col.name)
        self.assertEqual(('/redfish/v1/Systems/437XR1138R2/Processors/CPU1',
                          '/redfish/v1/Systems/437XR1138R2/Processors/CPU2'),
                         self.sys_processor_col.members_identities)

    @mock.patch.object(processor, 'Processor', autospec=True)
    def test_get_member(self, mock_processor):
        self.sys_processor_col.get_member(
            '/redfish/v1/Systems/437XR1138R2/Processors/CPU1')
        mock_processor.assert_called_once_with(
            self.sys_processor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Processors/CPU1',
            self.sys_processor_col.redfish_version, None)

    @mock.patch.object(processor, 'Processor', autospec=True)
    def test_get_members(self, mock_processor):
        members = self.sys_processor_col.get_members()
        calls = [
            mock.call(self.sys_processor_col._conn,
                      '/redfish/v1/Systems/437XR1138R2/Processors/CPU1',
                      self.sys_processor_col.redfish_version, None),
            mock.call(self.sys_processor_col._conn,
                      '/redfish/v1/Systems/437XR1138R2/Processors/CPU2',
                      self.sys_processor_col.redfish_version, None)
        ]
        mock_processor.assert_has_calls(calls)
        self.assertIsInstance(members, list)
        self.assertEqual(2, len(members))

    def _setUp_processor_summary(self):
        self.conn.get.return_value.json.reset_mock()
        successive_return_values = []
        file_names = ['sushy/tests/unit/json_samples/processor.json',
                      'sushy/tests/unit/json_samples/processor2.json']
        for file_name in file_names:
            with open(file_name) as f:
                successive_return_values.append(json.load(f))

        self.conn.get.return_value.json.side_effect = successive_return_values

    def test_summary(self):
        # | GIVEN |
        self._setUp_processor_summary()
        # | WHEN |
        actual_summary = self.sys_processor_col.summary
        # | THEN |
        self.assertEqual((16, sushy.PROCESSOR_ARCH_x86),
                         actual_summary)
        self.assertEqual(16, actual_summary.count)
        self.assertEqual(sushy.PROCESSOR_ARCH_x86,
                         actual_summary.architecture)

        # reset mock
        self.conn.get.return_value.json.reset_mock()

        # | WHEN & THEN |
        # tests for same object on invoking subsequently
        self.assertIs(actual_summary,
                      self.sys_processor_col.summary)
        self.conn.get.return_value.json.assert_not_called()

    def test_summary_on_refresh(self):
        # | GIVEN |
        self._setUp_processor_summary()
        # | WHEN & THEN |
        self.assertEqual((16, sushy.PROCESSOR_ARCH_x86),
                         self.sys_processor_col.summary)

        self.conn.get.return_value.json.side_effect = None
        # On refreshing the sys_processor_col instance...
        with open('sushy/tests/unit/json_samples/'
                  'processor_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        self.sys_processor_col.invalidate()
        self.sys_processor_col.refresh(force=False)

        # | GIVEN |
        self._setUp_processor_summary()
        # | WHEN & THEN |
        self.assertEqual((16, sushy.PROCESSOR_ARCH_x86),
                         self.sys_processor_col.summary)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# This is referred from Redfish standard schema.
# https://redfish.dmtf.org/schemas/Endpoint.v1_3_0.json

import logging

from sushy.resources import base
from sushy.resources import common
from sushy.resources.fabric import mappings as fab_maps
from sushy.resources import mappings as res_maps
from sushy import utils

LOG = logging.getLogger(__name__)


class IPv4AddressField(base.CompositeField):

    address = base.Field('Address')
    """This is the IPv4 Address."""

    gateway = base.Field('Gateway')
    """This is the IPv4 gateway for this address."""

    subnet_mask = base.Field('SubnetMask')
    """This is the IPv4 Subnet mask."""

    address_origin = base.MappedField('AddressOrigin',
                                      fab_maps.ADDRESS_ORIGIN_IPv4_VALUE_MAP)
    """This indicates how the address was determined."""


class IPv6AddressField(base.CompositeField):

    address = base.Field('Address')
    """This is the IPv6 Address."""

    prefix_length = base.Field('PrefixLength', adapter=utils.int_or_none)
    """This is the IPv6 Address Prefix Length."""

    address_origin = base.MappedField('AddressOrigin',
                                      fab_maps.ADDRESS_ORIGIN_IPv6_VALUE_MAP)
    """This indicates how the address was determined."""

    address_state = base.MappedField('AddressState',
                                     fab_maps.ADDRESS_STATE_VALUE_MAP)
    """The current state of this address as defined in RFC 4862."""


class IPTransportDetailsListField(base.ListField):
    """IP transport details

    This array contains details for each IP transport supported by this
    endpoint. The array structure can be used to model multiple IP addresses
    for this endpoint.
    """

    port = base.Field('Port', adapter=utils.int_or_none)
    """The UDP or TCP port number used by the Endpoint."""

    transport_protocol = base.MappedField('TransportProtocol',
                                          res_maps.PROTOCOL_TYPE_VALUE_MAP)
    """The protocol used by the connection entity."""

    ipv4_address = IPv4AddressField('IPv4Address')
    """The IPv4 address object."""

    ipv6_address = IPv6AddressField('IPv6Address')
    """The IPv6 address object."""


class PciIdField(base.CompositeField):

    device_id = base.Field('DeviceId')
    """The Device ID of this PCIe function."""

    subsystem_id = base.Field('SubsystemId')
    """The Subsystem ID of this PCIefunction."""

    subsystem_vendor_id = base.Field('SubsystemVendorId')
    """The Subsystem Vendor ID of thisPCIe function."""

    vendor_id = base.Field('VendorId')
    """The Vendor ID of this PCIe function."""


class ConnectedEntitiesListField(base.ListField):
    """All the entities connected to this endpoint."""

    pci_class_code = base.Field('PciClassCode')
    """The Class Code, Subclass code, and Programming Interface code of
    this PCIe function."""

    pci_function_number = base.Field('PciFunctionNumber',
                                     adapter=utils.int_or_none)
    """The PCI ID of the connected entity."""

    entity_pci_id = PciIdField('EntityPciId')
    """The PCI ID of the connected entity."""

    identifiers = common.IdentifiersListField('Identifiers', default=[])
    """Identifiers for the remote entity."""

    entity_role = base.MappedField('EntityRole',
                                   fab_maps.ENTITY_ROLE_VALUE_MAP)
    """The role of the connected entity."""

    entity_type = base.MappedField('EntityType',
                                   fab_maps.ENTITY_TYPE_VALUE_MAP)
    """The type of the connected entity."""


class Endpoint(base.ResourceBase):
    """This class represents a fabric endpoint.

    It represents the properties of an entity that sends or receives protocol
    defined messages over a transport.
    """

    identity = base.Field('Id', required=True)
    """Identifier for the endpoint"""

    name = base.Field('Name', required=True)
    """The endpoint name"""

    description = base.Field('Description')
    """The endpoint description"""

    status = common.StatusField('Status')
    """The endpoint status"""

    host_reservation_memory_bytes = base.Field('HostReservationMemoryBytes',
                                               adapter=utils.int_or_none)
    """The amount of memory in Bytes that the Host should allocate to connect
    to this endpoint.
    """

    endpoint_protocol = base.MappedField('EndpointProtocol',
                                         res_maps.PROTOCOL_TYPE_VALUE_MAP)
    """The protocol supported by this endpoint."""

    pci_id = PciIdField('PciId')
    """The PCI ID of the endpoint."""

    IP_transport_details = IPTransportDetailsListField('IPTransportDetails')
    """This array contains details for each IP transport supported by this
    endpoint. The array structure can be used to model multiple IP addresses
    for this endpoint."""

    connected_entities = ConnectedEntitiesListField('ConnectedEntities')
    """All entities connected to this endpoint."""


class EndpointCollection(base.ResourceCollectionBase):
    """Represents a collection of endpoints associated with the fabric."""

    @property
    def _resource_type(self):
        return Endpoint



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sushy.resources.chassis.power import constants as pow_cons

POWER_SUPPLY_TYPE_MAP = {
    'Unknown': pow_cons.POWER_SUPPLY_TYPE_UNKNOWN,
    'AC': pow_cons.POWER_SUPPLY_TYPE_AC,
    'DC': pow_cons.POWER_SUPPLY_TYPE_DC,
    'ACorDC': pow_cons.POWER_SUPPLY_TYPE_ACDC,
}

POWER_SUPPLY_INPUT_TYPE_MAP = {
    'AC': pow_cons.INPUT_TYPE_AC,
    'DC': pow_cons.INPUT_TYPE_DC,
}

LINE_INPUT_VOLTAGE_TYPE_MAP = {
    'Unknown': pow_cons.LINE_INPUT_VOLTAGE_TYPE_UNKNOWN,
    'ACLowLine': pow_cons.LINE_INPUT_VOLTAGE_TYPE_ACLOW,
    'ACMidLine': pow_cons.LINE_INPUT_VOLTAGE_TYPE_ACMID,
    'ACHighLine': pow_cons.LINE_INPUT_VOLTAGE_TYPE_ACHIGH,
    'DCNeg48V': pow_cons.LINE_INPUT_VOLTAGE_TYPE_DCNEG48,
    'DC380V': pow_cons.LINE_INPUT_VOLTAGE_TYPE_DC380,
    'AC120V': pow_cons.LINE_INPUT_VOLTAGE_TYPE_AC120,
    'AC240V': pow_cons.LINE_INPUT_VOLTAGE_TYPE_AC240,
    'AC277V': pow_cons.LINE_INPUT_VOLTAGE_TYPE_AC277,
    'ACandDCWideRange': pow_cons.LINE_INPUT_VOLTAGE_TYPE_ACDCWIDE,
    'ACWideRange': pow_cons.LINE_INPUT_VOLTAGE_TYPE_ACWIDE,
    'DC240V': pow_cons.LINE_INPUT_VOLTAGE_TYPE_DC240,
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Volume Initialization Types
VOLUME_INIT_TYPE_FAST = 'fast'
"""The volume is prepared for use quickly, typically by erasing just the
beginning and end of the space so that partitioning can be performed."""

VOLUME_INIT_TYPE_SLOW = 'slow'
"""The volume is prepared for use slowly, typically by completely erasing
the volume."""

# VolumeType Types
VOLUME_TYPE_RAW_DEVICE = 'rawdevice'
"""The volume is a raw physical device without any RAID or other
virtualization applied."""

VOLUME_TYPE_NON_REDUNDANT = 'nonredundant'
"""The volume is a non-redundant storage device."""

VOLUME_TYPE_MIRRORED = 'mirrored'
"""The volume is a mirrored device."""

VOLUME_TYPE_STRIPED_WITH_PARITY = 'stripedwithparity'
"""The volume is a device which uses parity to retain redundant information."""

VOLUME_TYPE_SPANNED_MIRRORS = 'spannedmirrors'
"""The volume is a spanned set of mirrored devices."""

VOLUME_TYPE_SPANNED_STRIPES_WITH_PARITY = 'spannedstripeswithparity'
"""The volume is a spanned set of devices which uses parity to retain
redundant information."""

# RAIDType Types
RAID_TYPE_RAID0 = 'RAID0'
"""A placement policy where consecutive logical blocks of data are uniformly
distributed across a set of independent storage devices without offering any
form of redundancy."""

RAID_TYPE_RAID1 = 'RAID1'
"""A placement policy where each logical block of data is stored on more than
one independent storage device."""

RAID_TYPE_RAID3 = 'RAID3'
"""A placement policy using parity-based protection where logical bytes of
data are uniformly distributed across a set of independent storage devices and
where the parity is stored on a dedicated independent storage device."""

RAID_TYPE_RAID4 = 'RAID4'
"""A placement policy using parity-based protection where logical blocks of
data are uniformly distributed across a set of independent storage devices and
where the parity is stored on a dedicated independent storage device."""

RAID_TYPE_RAID5 = 'RAID5'
"""A placement policy using parity-based protection for storing stripes of 'n'
logical blocks of data and one logical block of parity across a set of 'n+1'
independent storage devices where the parity and data blocks are interleaved
across the storage devices."""

RAID_TYPE_RAID6 = 'RAID6'
"""A placement policy using parity-based protection for storing stripes of 'n'
logical blocks of data and two logical blocks of independent parity across a
set of 'n+2' independent storage devices where the parity and data blocks are
interleaved across the storage devices."""

RAID_TYPE_RAID10 = 'RAID10'
"""A placement policy that creates a striped device (RAID 0) over a set of
mirrored devices (RAID 1)."""

RAID_TYPE_RAID01 = 'RAID01'
"""A data placement policy that creates a mirrored device (RAID 1) over a set
of striped devices (RAID 0)."""

RAID_TYPE_RAID6TP = 'RAID6TP'
"""A placement policy that uses parity-based protection for storing stripes of
'n' logical blocks of data and three logical blocks of independent parity
across a set of 'n+3' independent storage devices where the parity and data
blocks are interleaved across the storage devices."""

RAID_TYPE_RAID1E = 'RAID1E'
"""A placement policy that uses a form of mirroring implemented over a set of
independent storage devices where logical blocks are duplicated on a pair of
independent storage devices so that data is uniformly distributed across the
storage devices."""

RAID_TYPE_RAID50 = 'RAID50'
"""A placement policy that uses a RAID 0 stripe set over two or more RAID 5
sets of independent storage devices."""

RAID_TYPE_RAID60 = 'RAID60'
"""A placement policy that uses a RAID 0 stripe set over two or more RAID 6
sets of independent storage devices."""

RAID_TYPE_RAID00 = 'RAID00'
"""A placement policy that creates a RAID 0 stripe set over two or more RAID 0
sets."""

RAID_TYPE_RAID10E = 'RAID10E'
"""A placement policy that uses a RAID 0 stripe set over two or more RAID 10
sets."""

RAID_TYPE_RAID1Triple = 'RAID1Triple'
"""A placement policy where each logical block of data is mirrored three times
across a set of three independent storage devices."""

RAID_TYPE_RAID10Triple = 'RAID10Triple'
"""A placement policy that uses a striped device (RAID 0) over a set of triple
mirrored devices (RAID 1Triple)."""



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from unittest import mock

from sushy.resources.compositionservice import compositionservice
from sushy.resources.compositionservice import resourceblock
from sushy.resources.compositionservice import resourcezone
from sushy.resources import constants as res_cons
from sushy.tests.unit import base


class CompositionServiceTestCase(base.TestCase):

    def setUp(self):
        super(CompositionServiceTestCase, self).setUp()
        self.conn = mock.Mock()
        with open(
            'sushy/tests/unit/json_samples/compositionservice.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.comp_ser = compositionservice.CompositionService(
            self.conn,
            '/redfish/v1/CompositionService',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.comp_ser._parse_attributes(self.json_doc)
        self.assertFalse(self.comp_ser.allow_overprovisioning)
        self.assertTrue(self.comp_ser.allow_zone_affinity)
        self.assertTrue(self.comp_ser.description, 'CompositionService1')
        self.assertEqual(
            'CompositionService',
            self.comp_ser.identity)
        self.assertEqual(
            'Composition Service',
            self.comp_ser.name)
        self.assertEqual(res_cons.STATE_ENABLED, self.comp_ser.status.state)
        self.assertEqual(res_cons.HEALTH_OK, self.comp_ser.status.health)
        self.assertTrue(self.comp_ser.service_enabled)

    @mock.patch.object(resourceblock, 'ResourceBlockCollection', autospec=True)
    def test_get_resource_blocks(self, mock_resourceblock_col):
        _ = self.comp_ser.resource_blocks
        mock_resourceblock_col.assert_called_once_with(
            self.comp_ser._conn,
            self.comp_ser._get_resource_blocks_collection_path,
            self.comp_ser.redfish_version, None)

    @mock.patch.object(resourcezone, 'ResourceZoneCollection', autospec=True)
    def test_get_resource_zones(self, mock_resourcezone_col):
        _ = self.comp_ser.resource_zones
        mock_resourcezone_col.assert_called_once_with(
            self.comp_ser._conn,
            self.comp_ser._get_resource_zones_collection_path,
            self.comp_ser.redfish_version, None)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
from unittest import mock


import sushy
from sushy.resources.system.storage import drive
from sushy.resources.system.storage import storage
from sushy.resources.system.storage import volume
from sushy.tests.unit import base


STORAGE_DRIVE_FILE_NAMES = [
    'sushy/tests/unit/json_samples/drive.json',
    'sushy/tests/unit/json_samples/drive2.json',
    'sushy/tests/unit/json_samples/drive3.json'
]

STORAGE_VOLUME_FILE_NAMES = [
    'sushy/tests/unit/json_samples/volume_collection.json',
    'sushy/tests/unit/json_samples/volume.json',
    'sushy/tests/unit/json_samples/volume2.json',
    'sushy/tests/unit/json_samples/volume3.json'
]


class StorageTestCase(base.TestCase):

    def setUp(self):
        super(StorageTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.storage = storage.Storage(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Storage/1',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.storage._parse_attributes(self.json_doc)
        self.assertEqual('1.0.2', self.storage.redfish_version)
        self.assertEqual('1', self.storage.identity)
        self.assertEqual('Local Storage Controller', self.storage.name)
        self.assertEqual('ok', self.storage.status.health)
        self.assertEqual('ok', self.storage.status.health_rollup)
        self.assertEqual('enabled', self.storage.status.state)
        self.assertEqual(
            ('/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/35D38F11ACEF7BD3',  # noqa
             '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3F5A8C54207B7233',  # noqa
             '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/32ADF365C6C1B7BD',  # noqa
             '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3D58ECBC375FD9F2',  # noqa
             ), self.storage.drives_identities)

    def test_get_drive(self):
        # | WHEN |
        actual_drive = self.storage.get_drive(
            '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/'
            '35D38F11ACEF7BD3')
        # | THEN |
        self.assertIsInstance(actual_drive, drive.Drive)
        self.assertTrue(self.conn.get.return_value.json.called)

    @mock.patch.object(drive, 'Drive', autospec=True)
    def test_drives(self, Drive_mock):
        # | WHEN |
        all_drives = self.storage.drives
        # | THEN |
        calls = [
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/35D38F11ACEF7BD3',  # noqa
                      self.storage.redfish_version, None),
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3F5A8C54207B7233',  # noqa
                      self.storage.redfish_version, None),
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/32ADF365C6C1B7BD',  # noqa
                      self.storage.redfish_version, None),
            mock.call(self.storage._conn,
                      '/redfish/v1/Systems/437XR1138R2/Storage/1/Drives/3D58ECBC375FD9F2',  # noqa
                      self.storage.redfish_version, None)
        ]
        Drive_mock.assert_has_calls(calls)
        self.assertIsInstance(all_drives, list)
        self.assertEqual(4, len(all_drives))
        self.assertIsInstance(all_drives[0], drive.Drive.__class__)

        # returning cached value
        Drive_mock.reset_mock()
        # | WHEN |
        all_drives = self.storage.drives
        # | THEN |
        self.assertFalse(Drive_mock.called)
        self.assertIsInstance(all_drives, list)
        self.assertEqual(4, len(all_drives))
        self.assertIsInstance(all_drives[0], drive.Drive.__class__)

    def test_storage_controllers(self):
        controllers = self.storage.storage_controllers
        self.assertIsInstance(controllers, list)
        self.assertEqual(1, len(controllers))
        controller = controllers[0]
        self.assertEqual('0', controller.member_id)
        self.assertEqual('Contoso Integrated RAID', controller.name)
        self.assertEqual('ok', controller.status.health)
        self.assertEqual('enabled', controller.status.state)
        identifiers = controller.identifiers
        self.assertIsInstance(identifiers, list)
        self.assertEqual(1, len(identifiers))
        identifier = identifiers[0]
        self.assertEqual(sushy.DURABLE_NAME_FORMAT_NAA,
                         identifier.durable_name_format)
        self.assertEqual('345C59DBD970859C', identifier.durable_name)
        self.assertEqual(12, controller.speed_gbps)
        self.assertEqual([sushy.PROTOCOL_TYPE_PCIe],
                         controller.controller_protocols)
        self.assertEqual([sushy.PROTOCOL_TYPE_SAS, sushy.PROTOCOL_TYPE_SATA],
                         controller.device_protocols)
        self.assertEqual([sushy.RAID_TYPE_RAID0, sushy.RAID_TYPE_RAID1],
                         controller.raid_types)

    def test_drives_after_refresh(self):
        self.storage.refresh()
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        all_drives = self.storage.drives
        self.assertIsInstance(all_drives, list)
        self.assertEqual(4, len(all_drives))
        for drv in all_drives:
            self.assertIsInstance(drv, drive.Drive)

    def test_drives_max_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.storage.drives_max_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(899527000000, self.storage.drives_max_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_drives_max_size_bytes_after_refresh(self):
        self.storage.refresh()
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.storage.drives_max_size_bytes)

    def test_volumes(self):
        # | GIVEN |
        self.conn.get.return_value.json.reset_mock()
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # | WHEN |
        actual_volumes = self.storage.volumes
        # | THEN |
        self.assertIsInstance(actual_volumes,
                              volume.VolumeCollection)
        self.conn.get.return_value.json.assert_called_once_with()

    def test_volumes_cached(self):
        # | GIVEN |
        self.conn.get.return_value.json.reset_mock()
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # invoke it once
        actual_volumes = self.storage.volumes
        self.conn.get.return_value.json.reset_mock()
        # | WHEN & THEN |
        # tests for same object on invoking subsequently
        self.assertIs(actual_volumes,
                      self.storage.volumes)
        self.conn.get.return_value.json.assert_not_called()

    def test_volumes_on_refresh(self):
        # | GIVEN |
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # | WHEN & THEN |
        vols = self.storage.volumes
        self.assertIsInstance(vols, volume.VolumeCollection)

        # On refreshing the system instance...
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)

        self.storage.invalidate()
        self.storage.refresh(force=False)

        # | WHEN & THEN |
        self.assertTrue(vols._is_stale)

        # | GIVEN |
        with open('sushy/tests/unit/json_samples/volume_collection.json') as f:
            self.conn.get.return_value.json.return_value = json.load(f)
        # | WHEN & THEN |
        self.assertIsInstance(self.storage.volumes,
                              volume.VolumeCollection)


class StorageCollectionTestCase(base.TestCase):

    def setUp(self):
        super(StorageCollectionTestCase, self).setUp()
        self.conn = mock.Mock()
        with open('sushy/tests/unit/json_samples/'
                  'storage_collection.json') as f:
            self.json_doc = json.load(f)

        self.conn.get.return_value.json.return_value = self.json_doc

        self.stor_col = storage.StorageCollection(
            self.conn, '/redfish/v1/Systems/437XR1138R2/Storage',
            redfish_version='1.0.2')

    def test__parse_attributes(self):
        self.stor_col._parse_attributes(self.json_doc)
        self.assertEqual((
            '/redfish/v1/Systems/437XR1138R2/Storage/1',),
            self.stor_col.members_identities)

    @mock.patch.object(storage, 'Storage', autospec=True)
    def test_get_member(self, Storage_mock):
        self.stor_col.get_member(
            '/redfish/v1/Systems/437XR1138R2/Storage/1')
        Storage_mock.assert_called_once_with(
            self.stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1',
            self.stor_col.redfish_version, None)

    @mock.patch.object(storage, 'Storage', autospec=True)
    def test_get_members(self, Storage_mock):
        members = self.stor_col.get_members()
        Storage_mock.assert_called_once_with(
            self.stor_col._conn,
            '/redfish/v1/Systems/437XR1138R2/Storage/1',
            self.stor_col.redfish_version, None)
        self.assertIsInstance(members, list)
        self.assertEqual(1, len(members))

    def test_drives_sizes_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual([899527000000, 899527000000, 899527000000,
                          899527000000], self.stor_col.drives_sizes_bytes)

    def test_max_drive_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.stor_col.max_drive_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(899527000000, self.stor_col.max_drive_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_max_drive_size_bytes_after_refresh(self):
        self.stor_col.refresh(force=False)
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_DRIVE_FILE_NAMES + [STORAGE_DRIVE_FILE_NAMES[-1]]:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(899527000000, self.stor_col.max_drive_size_bytes)

    def test_volumes_sizes_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_VOLUME_FILE_NAMES:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual([107374182400, 899527000000, 1073741824000],
                         self.stor_col.volumes_sizes_bytes)

    def test_max_volume_size_bytes(self):
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_VOLUME_FILE_NAMES:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(1073741824000, self.stor_col.max_volume_size_bytes)

        # for any subsequent fetching it gets it from the cached value
        self.conn.get.return_value.json.reset_mock()
        self.assertEqual(1073741824000, self.stor_col.max_volume_size_bytes)
        self.conn.get.return_value.json.assert_not_called()

    def test_max_volume_size_bytes_after_refresh(self):
        self.stor_col.refresh(force=False)
        self.conn.get.return_value.json.reset_mock()

        successive_return_values = []
        with open('sushy/tests/unit/json_samples/storage.json') as f:
            successive_return_values.append(json.load(f))
        # repeating the 3rd one to provide mock data for 4th iteration.
        for fname in STORAGE_VOLUME_FILE_NAMES:
            with open(fname) as f:
                successive_return_values.append(json.load(f))
        self.conn.get.return_value.json.side_effect = successive_return_values

        self.assertEqual(1073741824000, self.stor_col.max_volume_size_bytes)



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from http import client as http_client
import logging
import re
import time
from urllib import parse as urlparse

import requests

from sushy import exceptions
from sushy.taskmonitor import TaskMonitor
from sushy import utils

LOG = logging.getLogger(__name__)


_SERVER_SIDE_RETRIES = 5
_SERVER_SIDE_RETRY_DELAY = 3


class Connector(object):

    def __init__(
            self, url, username=None, password=None, verify=True,
            response_callback=None):
        self._url = url
        self._verify = verify
        self._session = requests.Session()
        self._session.verify = self._verify
        self._response_callback = response_callback
        self._auth = None

        # NOTE(TheJulia): In order to help prevent recursive post operations
        # by allowing us to understand that we should stop authentication.
        self._sessions_uri = None
        # NOTE(etingof): field studies reveal that some BMCs choke at
        # long-running persistent HTTP connections (or TCP connections).
        # By default, we ask HTTP server to shut down HTTP connection we've
        # just used.
        self._session.headers['Connection'] = 'close'
        # NOTE(TheJulia): Depending on the BMC, offering compression as an
        # acceptable response changes the ETag behavior to offering an
        # automatic "weak" ETag response, which is appropriate because the
        # body content *may* not be a byte for byte match given compression.
        # Overall, the value of compression is less than the value of concise
        # interaction with the BMC. Setting to identity basically means
        # "without modification or compression". By default, python-requests
        # indicates responses can be compressed.
        self._session.headers['Accept-Encoding'] = 'identity'

        if username or password:
            LOG.warning('Passing username and password to Connector is '
                        'deprecated. Authentication is passed through '
                        'set_auth now, support for these arguments will '
                        'be removed in the future')
            self.set_http_basic_auth(username, password)

    def set_auth(self, auth):
        """Sets the authentication mechanism for our connector."""
        self._auth = auth

    def set_http_basic_auth(self, username, password):
        """Sets the http basic authentication information."""
        self._session.auth = (username, password)

    def set_http_session_auth(self, session_auth_token):
        """Sets the session authentication information."""
        self._session.auth = None
        self._session.headers.update({'X-Auth-Token': session_auth_token})

    def close(self):
        """Close this connector and the associated HTTP session."""
        self._session.close()

    def _op(self, method, path='', data=None, headers=None, blocking=False,
            timeout=60, server_side_retries=_SERVER_SIDE_RETRIES,
            **extra_session_req_kwargs):
        """Generic RESTful request handler.

        :param method: The HTTP method to be used, e.g: GET, POST,
            PUT, PATCH, etc...
        :param path: The sub-URI or absolute URL path to the resource.
        :param data: Optional JSON data.
        :param headers: Optional dictionary of headers.
        :param blocking: Whether to block for asynchronous operations.
        :param timeout: Max time in seconds to wait for blocking async call.
        :param extra_session_req_kwargs: Optional keyword argument to pass
         requests library arguments which would pass on to requests session
         object.
        :returns: The response object from the requests library.
        :raises: ConnectionError
        :raises: HTTPError
        """
        url = path if urlparse.urlparse(path).netloc else urlparse.urljoin(
            self._url, path)
        headers = headers or {}
        lc_headers = [k.lower() for k in headers]
        if data is not None and 'content-type' not in lc_headers:
            headers['Content-Type'] = 'application/json'
        if 'odata-version' not in lc_headers:
            headers['OData-Version'] = '4.0'
        # TODO(lucasagomes): We should mask the data to remove sensitive
        # information
        LOG.debug('HTTP request: %(method)s %(url)s; headers: %(headers)s; '
                  'body: %(data)s; blocking: %(blocking)s; timeout: '
                  '%(timeout)s; session arguments: %(session)s;',
                  {'method': method, 'url': url,
                   'headers': utils.sanitize(headers),
                   'data': utils.sanitize(data),
                   'blocking': blocking, 'timeout': timeout,
                   'session': extra_session_req_kwargs})
        try:
            response = self._session.request(method, url, json=data,
                                             headers=headers,
                                             **extra_session_req_kwargs)
        except requests.exceptions.RequestException as e:
            # Capture any general exception by looking for the parent
            # class of exceptions in the requests library.
            # Specifically this will cover cases such as transport
            # failures, connection timeouts, and encoding errors.
            # Raising this as sushy ConnectionError allows users to
            # understand something bad has happened, and to
            # allow them to respond accordingly.
            raise exceptions.ConnectionError(url=url, error=e)

        if self._response_callback:
            self._response_callback(response)

        # If we received an AccessError, and we
        # previously established a redfish session
        # there is a chance that the session has timed-out.
        # Attempt to re-establish a session.
        try:
            exceptions.raise_for_response(method, url, response)
        except exceptions.AccessError as e:
            if (method == 'POST'
                    and self._sessions_uri is not None
                    and self._sessions_uri in url):
                LOG.error('Authentication to the session service failed. '
                          'Please check credentials and try again.')
                raise
            if self._auth is not None:
                # self._session.auth value is only set when basic auth is used
                if self._session.auth is not None:
                    LOG.warning('We have encountered an AccessError when '
                                'using \'basic\' authentication. %(err)s',
                                {'err': str(e)})
                    # NOTE(TheJulia): There is no way to recover Basic auth,
                    # as we need the client to be re-launched with new
                    # credentials.
                    raise
                try:
                    if self._auth.can_refresh_session():
                        self._auth.refresh_session()
                    else:
                        LOG.warning('Session authentication appears to have '
                                    'been lost at some point in time. '
                                    'Connectivity may have been lost during '
                                    'a prior session refresh. Attempting to '
                                    're-authenticate.')
                        self._auth.authenticate()
                except exceptions.AccessError as refresh_exc:
                    LOG.error("A failure occured while attempting to refresh "
                              "the session. Error: %s", refresh_exc.message)
                    raise
                LOG.debug("Authentication refreshed successfully, "
                          "retrying the call.")
                try:
                    response = self._session.request(
                        method, url, json=data,
                        headers=headers,
                        **extra_session_req_kwargs)
                except exceptions.HTTPError as retry_exc:
                    LOG.error("Failure occured while attempting to retry "
                              "request after refreshing the session. Error: "
                              "%s", retry_exc.message)
                    raise
            else:
                if method == 'GET' and url.endswith('SessionService'):
                    LOG.debug('HTTP GET of SessionService failed %s, '
                              'this is expected prior to authentication',
                              e.message)
                else:
                    LOG.error("Authentication error detected. Cannot proceed: "
                              "%s", e.message)
                raise
        except exceptions.ServerSideError as e:
            if method.lower() != 'get' or server_side_retries <= 0:
                raise
            else:
                LOG.warning('Got server side error %s in response to a '
                            'GET request, retrying after %d seconds',
                            e, _SERVER_SIDE_RETRY_DELAY)
                time.sleep(_SERVER_SIDE_RETRY_DELAY)
                return self._op(method, path, data=data, headers=headers,
                                blocking=blocking, timeout=timeout,
                                server_side_retries=server_side_retries - 1,
                                **extra_session_req_kwargs)

        if blocking and response.status_code == 202:
            if not response.headers.get('Location'):
                m = ('HTTP response for %(method)s request to %(url)s '
                     'returned status 202, but no Location header'
                     % {'method': method, 'url': url})
                raise exceptions.ConnectionError(url=url, error=m)

            mon = TaskMonitor.from_response(self, response, path)
            mon.wait(timeout)
            response = mon.response
            exceptions.raise_for_response(method, url, response)

        LOG.debug('HTTP response for %(method)s %(url)s: '
                  'status code: %(code)s',
                  {'method': method, 'url': url,
                   'code': response.status_code})

        return response

    def get(self, path='', data=None, headers=None, blocking=False,
            timeout=60, **extra_session_req_kwargs):
        """HTTP GET method.

        :param path: Optional sub-URI path to the resource.
        :param data: Optional JSON data.
        :param headers: Optional dictionary of headers.
        :param blocking: Whether to block for asynchronous operations.
        :param timeout: Max time in seconds to wait for blocking async call.
        :param extra_session_req_kwargs: Optional keyword argument to pass
         requests library arguments which would pass on to requests session
         object.
        :returns: The response object from the requests library.
        :raises: ConnectionError
        :raises: HTTPError
        """
        return self._op('GET', path, data=data, headers=headers,
                        blocking=blocking, timeout=timeout,
                        **extra_session_req_kwargs)

    def post(self, path='', data=None, headers=None, blocking=False,
             timeout=60, **extra_session_req_kwargs):
        """HTTP POST method.

        :param path: Optional sub-URI path to the resource.
        :param data: Optional JSON data.
        :param headers: Optional dictionary of headers.
        :param blocking: Whether to block for asynchronous operations.
        :param timeout: Max time in seconds to wait for blocking async call.
        :param extra_session_req_kwargs: Optional keyword argument to pass
         requests library arguments which would pass on to requests session
         object.
        :returns: The response object from the requests library.
        :raises: ConnectionError
        :raises: HTTPError
        """
        return self._op('POST', path, data=data, headers=headers,
                        blocking=blocking, timeout=timeout,
                        **extra_session_req_kwargs)

    def _etag_handler(self, path='', data=None, headers=None, etag=None,
                      blocking=False, timeout=60, **extra_session_req_kwargs):
        """eTag handler containing workarounds for PATCH requests with eTags.

        :param path: Optional sub-URI path to the resource.
        :param data: Optional JSON data.
        :param headers: Optional dictionary of headers.
        :param etag: eTag string.
        :param blocking: Whether to block for asynchronous operations.
        :param timeout: Max time in seconds to wait for blocking async call.
        :param extra_session_req_kwargs: Optional keyword argument to pass
         requests library arguments which would pass on to requests session
         object.
        :returns: The response object from the requests library.
        :raises: ConnectionError
        :raises: HTTPError
        """
        if etag is not None:
            if not headers:
                headers = {}
            headers['If-Match'] = etag
        try:
            response = self._op('PATCH', path, data=data,
                                headers=headers,
                                blocking=blocking, timeout=timeout,
                                **extra_session_req_kwargs)
        except exceptions.HTTPError as resp:
            LOG.warning("Initial request with eTag failed: %s",
                        resp)
            if resp.status_code == http_client.PRECONDITION_FAILED:
                # NOTE(janders) if there was no eTag provided AND the response
                # is HTTP 412 Precondition Failed, raise the exception
                if not etag:
                    raise
                # checking for weak eTag
                pattern = re.compile(r'^(W\/)("\w*")$')
                match = pattern.match(etag)
                if match:
                    LOG.info("Weak eTag provided with original request to "
                             "%s. Attempting to conversion to strong eTag "
                             "and re-trying.", path)
                    # trying weak eTag converted to strong
                    headers['If-Match'] = match.group(2)
                    try:
                        response = self._op('PATCH', path, data=data,
                                            headers=headers,
                                            blocking=blocking,
                                            timeout=timeout,
                                            **extra_session_req_kwargs)
                    except exceptions.HTTPError as resp:
                        if (resp.status_code == http_client.
                           PRECONDITION_FAILED):
                            LOG.warning("Request to %s with weak eTag "
                                        "converted to strong eTag also "
                                        "failed. Making the final attempt "
                                        "with no eTag specified.", path)
                        response = None
                    if response:
                        return response
                else:
                    # eTag is strong, if it failed the only other thing
                    # to try is removing it entirely
                    # info
                    LOG.warning("Strong eTag provided - retrying request to "
                                "%s with eTag removed.", path)
                del headers['If-Match']
                try:
                    response = self._op('PATCH', path, data=data,
                                        headers=headers,
                                        blocking=blocking, timeout=timeout,
                                        **extra_session_req_kwargs)
                except exceptions.HTTPError as resp:
                    LOG.error("Final re-try with eTag removed has failed, "
                              "raising exception %s", resp)
                    raise
            else:
                raise
        return response

    def patch(self, path='', data=None, headers=None, etag=None,
              blocking=False, timeout=60, **extra_session_req_kwargs):
        """HTTP PATCH method.

        :param path: Optional sub-URI path to the resource.
        :param data: Optional JSON data.
        :param headers: Optional dictionary of headers.
        :param etag: Optional eTag string.
        :param blocking: Whether to block for asynchronous operations.
        :param timeout: Max time in seconds to wait for blocking async call.
        :param extra_session_req_kwargs: Optional keyword argument to pass
         requests library arguments which would pass on to requests session
         object.
        :returns: The response object from the requests library.
        :raises: ConnectionError
        :raises: HTTPError
        """
        return self._etag_handler(path, data,
                                  headers, etag,
                                  blocking, timeout,
                                  **extra_session_req_kwargs)

    def put(self, path='', data=None, headers=None, blocking=False,
            timeout=60, **extra_session_req_kwargs):
        """HTTP PUT method.

        :param path: Optional sub-URI path to the resource.
        :param data: Optional JSON data.
        :param headers: Optional dictionary of headers.
        :param blocking: Whether to block for asynchronous operations.
        :param timeout: Max time in seconds to wait for blocking async call.
        :param extra_session_req_kwargs: Optional keyword argument to pass
         requests library arguments which would pass on to requests session
         object.
        :returns: The response object from the requests library.
        :raises: ConnectionError
        :raises: HTTPError
        """
        return self._op('PUT', path, data=data, headers=headers,
                        blocking=blocking, timeout=timeout,
                        **extra_session_req_kwargs)

    def delete(self, path='', data=None, headers=None, blocking=False,
               timeout=60, **extra_session_req_kwargs):
        """HTTP DELETE method.

        :param path: Optional sub-URI path to the resource.
        :param data: Optional JSON data.
        :param headers: Optional dictionary of headers.
        :param blocking: Whether to block for asynchronous operations.
        :param timeout: Max time in seconds to wait for blocking async call.
        :param extra_session_req_kwargs: Optional keyword argument to pass
         requests library arguments which would pass on to requests session
         object.
        :returns: The response object from the requests library.
        :raises: ConnectionError
        :raises: HTTPError
        """
        return self._op('DELETE', path, data=data, headers=headers,
                        blocking=blocking, timeout=timeout,
                        **extra_session_req_kwargs)

    def __enter__(self):
        return self

    def __exit__(self, *_args):
        self.close()
